# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import os.path
import argparse

import _kht_constants as kht
import _kht_utils as utils
import _kht_families as test_families
import _kht_results_management as results

"""
Export de la sous-partie d'un repertoire LearningTest pour la famille de test specifiee
"""

# Liste des types d'export
EXPORT_ALL = "all"
EXPORT_REFERENCES = "references"
EXPORT_SCRIPTS = "scripts"
EXPORT_DATASETS = "datasets"
EXPORT_TYPES = [EXPORT_ALL, EXPORT_REFERENCES, EXPORT_SCRIPTS, EXPORT_DATASETS]
assert len(set(EXPORT_TYPES)) == len(EXPORT_TYPES), (
    "Export types " + str(EXPORT_TYPES) + " must not contain duplicates"
)

# Libelles des types d'exports
EXPORT_TYPE_LABELS = {
    EXPORT_ALL: "complete directory tree",
    EXPORT_REFERENCES: "only test dirs with the scripts and references results",
    EXPORT_SCRIPTS: "only test dirs with the scripts",
    EXPORT_DATASETS: "only datasets",
}
assert set(EXPORT_TYPE_LABELS) == set(
    EXPORT_TYPES
), "A label must be defined for each export type"


def export_learning_test_tree(
    home_dir,
    input_tool_dir_name,
    input_suite_dir_name,
    input_test_dir_name,
    destination,
    family,
    export_type,
):
    """
    Sauvegarde de l'arborescence de LearningTest
    Toute ou partie de l'arborescence source est sauvegardee selon la specification
     des operandes tool_dir_name, suite_dir_name, test_dir_name, qui peuvent etre None sinon.
    - home_dir: repertoire principal de l'aborescence source
    - tool_dir_name, suite_dir_name, test_dir_name: pour ne sauvegarder qu'une sous-partie
      de l'arborescence source si ces oprande ne sont pas None
    - destination: repertoire cible contenant l'arborescence sauvegardee
    - family: famille utilise pour choisir la sous-partie des suites a exporter
    - export_type: pour choisir une sous-partie de l'arborescence a exporter
    L'aborescence est sauvegardee dans <destination>/LearningTest avec _<export_type>
     en suffixe si lle type d'export est different de 'all'
    """
    assert os.path.isdir(home_dir)
    assert family in test_families.TEST_FAMILIES
    assert export_type in EXPORT_TYPES

    # Nom du repertoire cible
    target_home_dir = os.path.join(destination, kht.LEARNING_TEST)
    if export_type != EXPORT_ALL:
        target_home_dir += "_" + export_type

    # Verification qu'il n'existe pas deja
    if os.path.isdir(target_home_dir):
        utils.fatal_error("directory " + target_home_dir + " already exists")

    # Creation du repertoire cible
    try:
        os.makedirs(target_home_dir)
    except (IOError, os.error) as why:
        utils.fatal_error(
            "Unable to create target dir " + target_home_dir + " " + str(why)
        )

    # Parametrage des noms de fichiers ou repertoire specifiques a ne pas prendre en compte
    export_test_dirs = export_type in [EXPORT_ALL, EXPORT_SCRIPTS, EXPORT_REFERENCES]
    forbidden_names = ["__pycache__", kht.RESULTS, kht.COMPARISON_RESULTS_LOG]
    if export_type not in [EXPORT_ALL, EXPORT_REFERENCES]:
        # On ne sauvegarde pas les resultats de references dans tous leur contextes
        forbidden_names.append(kht.RESULTS_REF)

    # Initialisation des ensembles de dataset utilises par repertoire de jeux de donnees
    used_dataset_collections = {}
    for name in kht.DATASET_COLLECTION_NAMES:
        used_dataset_collections[name] = {}

    # Tous les outils sont a prendre en compte si on est a la racine
    if input_tool_dir_name is None:
        used_tool_names = kht.TOOL_NAMES
    # Sinon, seul l'outil correspondant au tool dir est a tester
    else:
        tool_name = kht.TOOL_NAMES_PER_DIR_NAME[input_tool_dir_name]
        used_tool_names = [tool_name]

    # Parcours des repertoires des outils pour en recopier les suites par famille
    exported_test_dir_number = 0
    for tool_name in used_tool_names:
        tool_dir_name = kht.TOOL_DIR_NAMES[tool_name]
        # Recherche des suites a utiliser
        if input_suite_dir_name is not None:
            assert tool_dir_name is not None
            suite_dir_names = [input_suite_dir_name]
        elif family == test_families.ALL:
            suite_dir_names = utils.sub_dirs(os.path.join(home_dir, tool_dir_name))
        else:
            suite_dir_names = test_families.FAMILY_TEST_SUITES[family, tool_name]
        # Chemin du repertoire de l'outil
        target_tool_dir = os.path.join(
            target_home_dir,
            tool_dir_name,
        )
        target_tool_dir_created = False
        # Parcours des suites de la famille
        for suite_dir_name in suite_dir_names:
            source_suite_dir = os.path.join(
                home_dir,
                tool_dir_name,
                suite_dir_name,
            )
            if os.path.isdir(source_suite_dir):
                # Chemin du repertoire de suite
                target_suite_dir = os.path.join(
                    target_home_dir,
                    tool_dir_name,
                    suite_dir_name,
                )
                target_suite_dir_created = False
                # Repertoires de test a utiliser
                if input_test_dir_name is not None:
                    test_dir_names = [input_test_dir_name]
                else:
                    test_dir_names = os.listdir(source_suite_dir)
                # Parcours des repertoires de test de la suite
                for test_dir_name in test_dir_names:
                    source_test_dir = os.path.join(source_suite_dir, test_dir_name)
                    # Export du contenu de la suite si necessaire
                    if export_test_dirs:
                        # Creation du repertoire de l'outil uniquement si necessaire
                        if not target_tool_dir_created:
                            utils.make_dir(target_tool_dir)
                            target_tool_dir_created = os.path.isdir(target_tool_dir)
                        # Creation du repertoire de suite uniquement si necessaire
                        if not target_suite_dir_created:
                            utils.make_dir(target_suite_dir)
                            target_suite_dir_created = os.path.isdir(target_suite_dir)
                        # Cas d'un repertoire
                        if os.path.isdir(source_test_dir):
                            # Copie du repertoire de test
                            target_test_dir = os.path.join(
                                target_suite_dir, test_dir_name
                            )
                            exported_test_dir_number += 1
                            copy_subdirs(
                                source_test_dir,
                                target_test_dir,
                                ignore_list=forbidden_names,
                            )
                        # Cas d'un fichier de la suite (ex: readme), a conserver
                        if os.path.isfile(source_test_dir):
                            source_file_name = source_test_dir
                            target_file_name = os.path.join(
                                target_suite_dir, test_dir_name
                            )
                            utils.copy_file(source_file_name, target_file_name)
                    # Gestion des repertoires de la suite a analyser
                    if os.path.isdir(source_test_dir):
                        # Analyse du scenario pour detecter les jeux de donnees utilises
                        test_prm_path = os.path.join(source_test_dir, kht.TEST_PRM)
                        if os.path.isfile(test_prm_path):
                            with open(
                                test_prm_path, "r", errors="ignore"
                            ) as test_prm_file:
                                lines = test_prm_file.readlines()
                                for line in lines:
                                    # Nettoyage de la ligne
                                    line = line.strip()
                                    pos_comment = line.find("//")
                                    if pos_comment >= 0:
                                        line = line[:pos_comment]
                                        line = line.strip()
                                    # Parsing de la ligne pour recherche les noms de dataset en derniere position
                                    # En effet, les patterns de type "ClassFileName ../../../datasets/Adult/Adult.kdic"
                                    # peuvent etre separes par un ou plusieurs caracteres blancs ou tabulation
                                    line = line.replace("\t", " ")
                                    fields = line.split(" ")
                                    if len(fields) >= 2:
                                        candidate_file_path = fields[-1]
                                        if candidate_file_path[0] == ".":
                                            # Le separateur est normalise dans les fichier scenario
                                            path_components = candidate_file_path.split(
                                                "/"
                                            )
                                            # les datasets se trouvent toujours au meme niveau relatif de l'arborescence
                                            if (
                                                len(path_components) > 5
                                                and path_components[0] == ".."
                                                and path_components[1] == ".."
                                                and path_components[2] == ".."
                                            ):
                                                dataset_collection = path_components[3]
                                                if (
                                                    dataset_collection
                                                    in kht.DATASET_COLLECTION_NAMES
                                                ):
                                                    dataset_name = path_components[4]
                                                    used_dataset_collections[
                                                        dataset_collection
                                                    ][dataset_name] = True
            else:
                print("error : suite directory not found: " + source_suite_dir)

    # Recopie des jeux de donnees selon les types d'export utilises si on exporte tout
    if export_type in [EXPORT_ALL, EXPORT_DATASETS]:
        # Recopie exhaustive des jeux de donnees utilises si on est a la racine avec toutes les suites
        if input_tool_dir_name is None and family == test_families.ALL:
            for dataset_collection in kht.DATASET_COLLECTION_NAMES:
                source_dataset_collection_dir = os.path.join(
                    home_dir, dataset_collection
                )
                target_dataset_collection_dir = os.path.join(
                    target_home_dir, dataset_collection
                )
                if os.path.isdir(source_dataset_collection_dir):
                    copy_subdirs(
                        source_dataset_collection_dir,
                        target_dataset_collection_dir,
                        ignore_list=forbidden_names,
                    )
        # Recopie des jeux de donnees utilises, sans selection sinon
        else:
            for dataset_collection in used_dataset_collections:
                used_dataset_names = used_dataset_collections[dataset_collection]
                if len(used_dataset_names):
                    # Creation du repertoire global de jeu de donnees
                    target_dataset_collection_dir = os.path.join(
                        target_home_dir,
                        dataset_collection,
                    )
                    utils.make_dir(target_dataset_collection_dir)
                    # Creation des repertoires elementaires de jeux de donnees
                    for dataset_name in used_dataset_names:
                        source_dataset_dir = os.path.join(
                            home_dir, dataset_collection, dataset_name
                        )
                        target_dataset_dir = os.path.join(
                            target_home_dir, dataset_collection, dataset_name
                        )
                        copy_subdirs(
                            source_dataset_dir,
                            target_dataset_dir,
                            ignore_list=forbidden_names,
                        )
    print("\tExported test dir number: " + str(exported_test_dir_number))


def copy_subdirs(source_dir, target_dir, ignore_list=None):
    """
    Copie recursive d'un repertoires source vers un repertoire cible
    :param source_dir:
    :param target_dir:
    :param ignore_list: nom des fichiers ou repertoires a ignorer de facon siliencieuse
    :return:
    """

    def is_dir_ignored(input_dir):
        """
        Verification si un repertoire peut etre ignorer, s'il est vide ou si son contenu peut eTre ignorer
        """
        input_names = os.listdir(input_dir)
        # Ignorer si le repertoire est vide
        if len(input_names) == 0:
            return True
        # Ne pas ignorer si non vide at iren a ignorer
        elif ignore_list is None and len(input_names) > 0:
            return False
        # Sinon, on parcours le repertoire pour voir s'il a un contenu a ignorer
        else:
            ignore_results = kht.RESULTS_REF in ignore_list
            for input_name in input_names:
                # Ne pas ignorer car le contenu n'est pas a ignorer
                if input_name not in ignore_list:
                    return False
                # Idem pour le cas particulier des resultst de reference
                elif ignore_results and not results.is_candidate_results_ref_dir(
                    input_name
                ):
                    return False
            # Peut etre ignorer si on a rien detecte
            return True

    # Retour si on peut ignorer le repertoire
    if is_dir_ignored(source_dir):
        return
    # Creation du repertoire cible
    if not os.path.isdir(target_dir):
        utils.make_dir(target_dir)
    # Parcours du repertoire source et propagation de la copie
    names = os.listdir(source_dir)
    # Copie du contenu
    ignore_reference_results = kht.RESULTS_REF in ignore_list
    for name in names:
        copy = True
        if ignore_list is not None:
            if name in ignore_list:
                copy = False
            elif ignore_reference_results and results.is_candidate_results_ref_dir(
                name
            ):
                copy = False
        if copy:
            source_path = os.path.join(source_dir, name)
            target_path = os.path.join(target_dir, name)
            if os.path.isdir(source_path):
                copy_subdirs(
                    source_path,
                    target_path,
                    ignore_list=ignore_list,
                )
            elif os.path.isfile(source_path):
                utils.copy_file(source_path, target_path)


def main():
    def build_usage_help(
        help_command,
        help_tool_dir_name=None,
        help_suite_dir_name=None,
        help_test_dir_name=None,
        help_options=None,
    ):
        """Construction d'une ligne d'aide pour un usage de la commande test"""
        source_dir = os.path.join(".", kht.LEARNING_TEST)
        if help_test_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name, help_test_dir_name
            )
        elif help_suite_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name
            )
        elif help_tool_dir_name is not None:
            source_dir = os.path.join(source_dir, help_tool_dir_name)
        usage_help = help_command + " " + source_dir + " (target dir)"
        if help_options is not None:
            usage_help += " " + help_options
        return usage_help

    # Nom du script
    script_file_name = os.path.basename(__file__)
    script_name = os.path.splitext(script_file_name)[0]

    # Ajout d'exemples d'utilisation
    epilog = ""
    epilog += "Usage examples"
    epilog += "\n  " + build_usage_help(script_name, help_options="-f all")
    epilog += "\n  " + build_usage_help(
        script_name, help_options="-f all --export-type references"
    )
    epilog += "\n  " + build_usage_help(
        script_name,
        kht.TOOL_DIR_NAMES[kht.KHIOPS],
        help_options="--export-type datasets",
    )
    epilog += "\n  " + build_usage_help(
        script_name,
        kht.TOOL_DIR_NAMES[kht.KHIOPS],
        "Standard",
        "Iris",
    )

    # Affichage de la liste des types d'exports, en la formattant au mieux
    export_types_help = ""
    max_export_type_len = 0
    for export_type in EXPORT_TYPES:
        max_export_type_len = max(max_export_type_len, len(export_type))
    for export_type in EXPORT_TYPES:
        label = EXPORT_TYPE_LABELS[export_type]
        export_types_help += "\n  " + export_type.ljust(max_export_type_len + 1) + label
        if export_type == EXPORT_ALL:
            export_types_help += " (default type)"

    # Parametrage de l'analyse de la ligne de commande
    parser = argparse.ArgumentParser(
        prog=script_name,
        description="export a subset of a source "
        + kht.LEARNING_TEST
        + " tree to a target dir named "
        + kht.LEARNING_TEST
        + " under the destination directory,"
        "\n the target directory is suffixed by _<type> if the export type is not 'all'",
        epilog=epilog,
        formatter_class=utils.get_formatter_class(script_name),
    )

    # Arguments positionnels
    utils.argument_parser_add_source_argument(parser)
    utils.argument_parser_add_dest_argument(parser)

    # Arguments optionnels standards
    utils.argument_parser_add_family_argument(parser)

    # Argument sur le types d'export
    parser.add_argument(
        "--export-type",
        help="export type" + export_types_help,
        choices=EXPORT_TYPES,
        default=EXPORT_ALL,
        metavar="type",
        action="store",
    )

    # Analyse de la ligne de commande
    args = parser.parse_args()

    # Verification de l'argument source
    (
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
    ) = utils.argument_parser_check_source_argument(parser, args.source)

    # Verification de l'argument destination
    utils.argument_parser_check_destination_dir(parser, home_dir, args.dest)

    # Lancement de la commande
    export_learning_test_tree(
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
        args.dest,
        args.family,
        args.export_type,
    )


if __name__ == "__main__":
    utils.set_flushed_outputs()
    main()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os


def Transcode(outputFileName, outputDataFileName):
    # ouverture des fichiers
    # fInput = open("splice.data", 'r')
    # avec les donnees sans doublons
    fInput = open("splice.data.cleaned", "r")
    fOutput = open(outputFileName, "w")
    fOutputData = open(outputDataFileName, "w")

    # ligne d'entete
    fOutput.write("SampleId\tClass\n")
    fOutputData.write("SampleId\tPos\tChar\n")

    # parcours de repertoires
    lines = fInput.readlines()
    for line in lines:
        line = line.replace("\n", "")
        line = line.replace(" ", "")
        fields = line.split(",")
        Class = fields[0]
        sampleId = fields[1]
        DNA = fields[2]
        fOutput.write(sampleId + "\t" + Class + "\n")
        index = 1
        for char in DNA:
            fOutputData.write(sampleId + "\t" + str(index) + "\t" + char + "\n")
            index = index + 1

    # fermeture  des fichiers
    fInput.close()
    fOutput.close()
    fOutputData.close()


Transcode("SpliceJunction.txt", "SpliceJunctionDNA.txt")



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import argparse

import _kht_constants as kht
import _kht_utils as utils


def print_env_var_help(env_var, help_text):
    """Affichage de l'aide sur une variable d'environnement"""
    print(env_var + ": " + str(os.getenv(env_var)) + "\n\t" + help_text)


def get_env_var_help_label(env_var, help_text):
    """Renvoie un libelle d'aide sur une variable d'environnement"""
    return env_var + ": " + str(os.getenv(env_var)) + ", " + help_text


def help_env_vars():
    # Aide sur les variables d'environnement influant le comportement des outils Khiops
    print("")
    print_env_var_help(
        kht.KHIOPS_PREPARATION_TRACE_MODE,
        "trace for dimensionnining of preparation tasks (default: false)",
    )
    print_env_var_help(kht.KHIOPS_PARALLEL_TRACE, "trace for parallel tasks (0 to 3)")
    print_env_var_help(
        kht.KHIOPS_FILE_SERVER_ACTIVATED, "activate MPI file server (false,true)"
    )

    # Aide particulier sur le pilotage des traces memoire
    print(
        "Analysis of memory stats"
        + "\n\t"
        + get_env_var_help_label(
            kht.KHIOPS_MEM_STATS_LOG_FILE_NAME, "memory stats log file name"
        )
        + "\n\t"
        + get_env_var_help_label(
            kht.KHIOPS_MEM_STATS_LOG_FREQUENCY,
            "frequency of allocator stats collection (0, 100000, 1000000,...)",
        )
        + "\n\t"
        + get_env_var_help_label(
            kht.KHIOPS_MEM_STATS_LOG_TO_COLLECT,
            "stats to collect (8193: only time and labels, 16383: all,...)",
        )
        + "\n\t"
        + get_env_var_help_label(
            kht.KHIOPS_IO_TRACE_MODE, "to collect IO trace (false, true)"
        )
    )


def main():
    """Fonction principale d'affichage de l'aide sur les variables d'environnement"""
    # Parametrage de l'analyse de la ligne de commande
    script_name = os.path.basename(__file__)
    base_script_name = os.path.splitext(script_name)[0]
    parser = argparse.ArgumentParser(
        prog=base_script_name,
        description="show the status of the main environment variables used by the tool binaries",
    )
    # Analyse de la ligne de commande et execution
    parser.parse_args()
    help_env_vars()


if __name__ == "__main__":
    utils.set_flushed_outputs()
    main()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os.path
import sys
import argparse

import _kht_constants as kht
import _kht_utils as utils
import _kht_families as test_families
import _kht_results_management as results
import _kht_standard_instructions as standard_instructions
import _kht_one_shot_instructions as one_shot_instructions

"""
Gestion de l'ensemble des instructions
"""


def apply_instruction_on_suite_dir(
    instruction_function,
    suite_dir,
    input_test_dir_name,
    min_test_time=None,
    max_test_time=None,
):
    """Application d'une instruction sur une suite ou sur un repertoire de test specifique"""
    assert suite_dir != ""

    # Erreur si repertoire de suite absent
    if not os.path.isdir(suite_dir):
        utils.fatal_error("missing directory for test suite " + suite_dir)

    # Collecte des sous-repertoires de test
    test_list = []
    # Cas avec un repertoire de test specifique
    if input_test_dir_name is not None:
        if os.path.isdir(os.path.join(suite_dir, input_test_dir_name)):
            test_list.append(input_test_dir_name)
    # Cas avec une suite de test
    else:
        for name in os.listdir(suite_dir):
            if os.path.isdir(os.path.join(suite_dir, name)):
                test_list.append(name)

    # Tri pour assurer la reproductibilite inter plateforme
    test_list.sort()

    # Execution de l'instruction
    for test_dir_name in test_list:
        # lanceur d'instruction sur un directory
        test_dir = os.path.realpath(os.path.join(suite_dir, test_dir_name))
        # verification de l'existence du directory
        if not os.path.isdir(test_dir):
            utils.fatal_error("directory " + test_dir + " does not exist")
        # On ne prend en compte que les tests compatibles avedc les contraintes de temps
        if results.is_results_ref_dir_time_selected(
            test_dir, min_test_time, max_test_time
        ):
            # Application de l'instruction
            current_dir = os.getcwd()
            os.chdir(test_dir)
            instruction_function(test_dir)
            os.chdir(current_dir)
    # Message synthetique de fin
    suite_dir_name = utils.dir_name(suite_dir)
    tool_dir_name = utils.parent_dir_name(suite_dir, 1)
    if input_test_dir_name is None:
        print("DONE\t" + tool_dir_name + "\t" + suite_dir_name)
    else:
        print(
            "done\t"
            + tool_dir_name
            + "\t"
            + suite_dir_name
            + "\t"
            + input_test_dir_name
        )


def apply_instruction_on_learning_test_tree(
    home_dir,
    input_tool_dir_name,
    input_suite_dir_name,
    input_test_dir_name,
    instruction_function,
    family,
    **kwargs
):
    """Applique une instruction un ensemble de suites de tests
    Toute ou partie de l'arborescence est prise en compte selon la specification
     des operandes tool_dir_name, suite_dir_name, test_dir_name, qui peuvent etre None sinon.
    - home_dir: repertoire principal de l'aborescence source
    - tool_dir_name, suite_dir_name, test_dir_name: pour ne prendre en compte qu'une sous-partie
      de l'arborescence source si ces oprande ne sont pas None
    - instruction_function: instruction a appliquee
    - family: famille utilise pour choisir la sous-partie des suites a exporter
    - kwargs: argument optionnels de la ligne de commande
    """

    # Tous les outils sont a prendre en compte si on est a la racine
    if input_tool_dir_name is None:
        used_tool_names = kht.TOOL_NAMES
    # Sinon, seul l'outil correspondant au tool dir est a tester
    else:
        tool_name = kht.TOOL_NAMES_PER_DIR_NAME[input_tool_dir_name]
        used_tool_names = [tool_name]

    # Cas d'un seul outil avec un repertoire de suite au de test specifique
    # Dans ce cas, on ignore la famille
    if input_suite_dir_name is not None:
        suite_dir = os.path.join(home_dir, input_tool_dir_name, input_suite_dir_name)
        apply_instruction_on_suite_dir(
            instruction_function, suite_dir, input_test_dir_name, **kwargs
        )
    # Cas d'un ou plusieurs outils, ou il faut utiliser les suites de la famille specifiee
    else:
        assert len(used_tool_names) >= 1
        for tool_name in used_tool_names:
            tool_dir_name = kht.TOOL_DIR_NAMES[tool_name]
            if family == test_families.ALL:
                test_suites = utils.sub_dirs(os.path.join(home_dir, tool_dir_name))
            else:
                test_suites = test_families.FAMILY_TEST_SUITES[family, tool_name]
            # Parcours de toutes les suites
            for name in test_suites:
                suite_dir = os.path.join(home_dir, tool_dir_name, name)
                if os.path.isdir(suite_dir):
                    apply_instruction_on_suite_dir(
                        instruction_function, suite_dir, None, **kwargs
                    )
                else:
                    print("error : suite directory not found: " + suite_dir)


def register_all_instructions():
    """Enregistrement de toutes les instructions standards et a usage unique
    Retourne un dictionnaire des instructions, et le nombre d'instructions standards
    """

    instructions1 = standard_instructions.register_standard_instructions()
    instructions2 = one_shot_instructions.register_one_shot_instructions()
    # L'operateur d'union entre dictionnaires '|' n'est supporte que depuis python 3.9
    all_instructions = instructions1
    all_instructions.update(instructions2)
    return all_instructions, len(instructions1)


def main():
    """Fonction principale d'application systematique d'une instruction sur une suite de test"""

    def build_usage_help(
        help_command,
        help_instruction,
        help_tool_dir_name=None,
        help_suite_dir_name=None,
        help_test_dir_name=None,
        help_options=None,
    ):
        """Construction d'une ligne d'aide pour un usage de la commande test"""
        source_dir = os.path.join(".", kht.LEARNING_TEST)
        if help_test_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name, help_test_dir_name
            )
        elif help_suite_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name
            )
        elif help_tool_dir_name is not None:
            source_dir = os.path.join(source_dir, help_tool_dir_name)
        usage_help = help_command + " " + source_dir + " " + help_instruction
        if help_options is not None:
            usage_help += " " + help_options
        return usage_help

    # Enregistrement de toutes les instructions
    (
        all_instructions,
        standard_instruction_number,
    ) = register_all_instructions()

    # Nom du script
    script_file_name = os.path.basename(__file__)
    script_name = os.path.splitext(script_file_name)[0]

    # Ajout d'exemples d'utilisation
    epilog = ""
    epilog += "Usage examples"
    epilog += "\n  " + build_usage_help(script_name, "errors")
    epilog += "\n  " + build_usage_help(
        script_name,
        "logs",
        kht.TOOL_DIR_NAMES[kht.KHIOPS],
        "Standard",
        "Iris",
    )
    epilog += "\n  " + build_usage_help(
        script_name,
        "errors",
        kht.TOOL_DIR_NAMES[kht.COCLUSTERING],
        help_options="-f basic",
    )

    # Affichage de la liste des instructions disponibles, en la formattant au mieux
    instructions_help = ""
    max_id_len = 0
    for instruction_id in all_instructions:
        max_id_len = max(max_id_len, len(instruction_id))
    for index, instruction_id in enumerate(all_instructions):
        (instruction_function, instruction_label) = all_instructions[instruction_id]
        if index == standard_instruction_number:
            instructions_help += "\none-shot instructions"
        instructions_help += (
            "\n  " + instruction_id.ljust(max_id_len + 1) + instruction_label
        )

    # Parametrage de l'analyse de la ligne de commande
    parser = argparse.ArgumentParser(
        prog=script_name,
        description="apply instruction (ex: errors) on a subset of test dirs",
        epilog=epilog,
        formatter_class=utils.get_formatter_class(script_name),
    )

    # Arguments positionnels
    utils.argument_parser_add_source_argument(parser)
    parser.add_argument(
        "instruction",
        help="instruction to apply" + instructions_help,
    )

    # Arguments optionnels standards
    utils.argument_parser_add_family_argument(parser)
    utils.argument_parser_add_processes_argument(parser)
    utils.argument_parser_add_forced_platform_argument(parser)
    utils.argument_parser_add_limit_test_time_arguments(parser)

    # Analyse de la ligne de commande
    args = parser.parse_args()

    # Verification de l'argument source
    (
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
    ) = utils.argument_parser_check_source_argument(parser, args.source)

    # Verification de l'argument instruction
    # On n'utilise pas le parametre 'choices' de add_argument pour eviter
    # d'avoir des messages d'erreur tres long comportant toutes les valeurs possibles
    if args.instruction not in all_instructions:
        parser.error(
            "argument instruction: unknown instruction '" + args.instruction + "'"
        )

    # Verification des arguments optionnels
    utils.argument_parser_check_processes_argument(parser, args.n)
    utils.argument_parser_check_limit_test_time_arguments(
        parser, args.min_test_time, args.max_test_time
    )

    # Memorisation des variables globales de gestion du contexte des resultats de reference
    results.process_number = args.n
    results.forced_platform = args.forced_platform

    # Acces a l'instruction a executer
    (instruction_function, instruction_label) = all_instructions[args.instruction]

    # Lancement de la commande
    apply_instruction_on_learning_test_tree(
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
        instruction_function,
        args.family,
        min_test_time=args.min_test_time,
        max_test_time=args.max_test_time,
    )


if __name__ == "__main__":
    utils.set_flushed_outputs()
    main()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

# Installation of prerequisites using pip
#  python -m pip install "ipywidgets>=7.2"
#  python -m pip install "notebook>=5.3"
#  python -m pip install pandas
#  python -m pip install plotly

# Documentation de plotly:
#  https://plotly.com/python/
#  https://plotly.com/python/getting-started/
#  https://plotly.com/python/reference/
#  https://plotly.com/python/text-and-annotations/

import random
import os
import re
from select import select

import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio
import numpy as np
import utils

pio.renderers.default = "browser"

"""
Visualizer for memory stats log produced by the Khiops tool

It takes log file as input and visualize them in a browser, with curves per memory stats and
summary bars per process.
"""


# Variables of Khiops memory log file, that can be choosen for visualisation
#   Time  is used for the X axis of the curves
#   Label is used to annotated somme position on the curves and process, to eaase the interpretation
#   Each other variable can be choosen to select curves to visualize
memory_stats_column_names = [
    "Time",
    "Heap mem",
    "Max heap mem",
    "Total heap mem",
    "Alloc",
    "Max alloc",
    "Total alloc",
    "Total free",
    "Granted",
    "Max granted",
    "Total requested size",
    "Total granted size",
    "Total free size",
    "Label",
]

# Specify possible views, with subset of variables selected for visualization
view_curves = {
    "Memory": ["Heap mem", "Granted"],
    "Allocs": ["Max alloc", "Alloc"],
    "Heap": ["Max heap mem", "Heap mem"],
    "TotalMemory": ["Total granted size", "Total free size"],
    "TotalAlloc": ["Total alloc", "Total free"],
    "TotalRequestedMemory": ["Total granted size", "Total requested size"],
}


def memory_stats_visualizer(
    file_name, current_view, fig_title=None, save_as_html=False
):
    """Visualize Khiops memory log file
    Parameters:
        file_name: name of the input memory stat log file, without suffix for the master process and
          with a suffix '_<processId>' per slave process
        current_view: chose among the variable or view names which curves to plot
        fig_title: specific figure title
        save_as_html: to save the figure as a html file, that can be used to launch
          the visualization without the log files
    """
    assert os.path.isfile(file_name), "Missing file " + file_name
    assert current_view in view_curves or current_view in memory_stats_column_names, (
        'View "' + current_view + '" is missing in available views or column names'
    )

    #################################################################################
    # Internal functions

    def add_bars(local_process_id, local_family, local_figure, local_process_legends):
        """Add bars on figure
        Input:
            local_process_id: id of process (0 for master)
            local_family: family of annotations
            local_figure: FIG_MEMORY_STATS, to display full bar on main figure
                    FIG_TASKS: to display task bars per process
            local_process_legends: dictionnaire des legendes deja affichees
        """
        x_ref = 0
        annotation_index = 0
        annotation_depth = 0
        bars_x = []
        bars_width = []
        bars_opacity = []
        bars_text = []
        bar_height_delta = 0.01
        if process_number == 1 or local_figure == FIG_MEMORY_STATS:
            local_name_suffix = ""
        else:
            local_name_suffix = (
                " Master"
                if local_process_id == 0
                else " (S" + str(local_process_id) + ")"
            )
        # Retrieve annotation prameters
        annotation_x = annotation_parameters[local_process_id, local_family][X]
        annotation_text = annotation_parameters[local_process_id, local_family][TEXT]

        # Analyse annnotations
        # Stack of annotations, according to the depth of the (Begin, End) couples
        hovertext_stack = []
        for i in range(len(annotation_x)):
            x = annotation_x[i]
            local_label = annotation_text[i]
            # Identify begin and end of task from labels
            is_begin = " Begin" in local_label or i == len(annotation_x) - 1
            is_end = " End" in local_label

            # Add bar per task and per inter-task
            if is_begin or is_end:
                # Increment annotation position
                if is_begin:
                    annotation_depth += 1
                else:
                    annotation_depth -= 1
                    assert annotation_depth >= 0, (
                        "Annotation depth below 0: " + local_label
                    )
                if is_end:
                    annotation_index += 1

                # Compute current hovertext
                hovertext = ""
                if (
                    local_figure == FIG_MEMORY_STATS
                ):  # Task index is added only for memory stats figure
                    hovertext = local_family + " " + str(annotation_index) + "<br>"
                # Clean Begin and End keywords
                if is_begin:
                    hovertext += local_label[: local_label.find(" Begin")]
                else:
                    hovertext += local_label[: local_label.find(" End")]
                # Add time and duration information
                hovertext += local_name_suffix
                hovertext += "<br>Start: " + "{:.3f}".format(x_ref)
                hovertext += "<br>Stop: " + "{:.3f}".format(x)
                hovertext += "<br>Duration: " + "{:.3f}".format(x - x_ref)

                # Specify the hovertext to use
                if is_end:
                    hovertext_stack.pop()  # Pop last hovertext
                else:
                    hovertext_stack.append(hovertext)  # Push new hovertext
                    # In case of nested annotation, use the preceding annotatation related to a Begin
                    if annotation_depth > 1:
                        hovertext = hovertext_stack[len(hovertext_stack) - 2]
                    # Add an inter-annoation label at the deapth 1
                    else:
                        hovertext = "Inter " + local_family
                        hovertext += "<br>Start: " + "{:.3f}".format(x_ref)
                        hovertext += "<br>Stop: " + "{:.3f}".format(x)
                        hovertext += "<br>Duration: " + "{:.3f}".format(x - x_ref)

                # Add bar spec
                if local_figure == FIG_MEMORY_STATS or is_end or annotation_depth > 1:
                    width = x - x_ref
                    bars_x.append(x_ref + width / 2)
                    bars_width.append(width)
                    bars_opacity.append(
                        0
                        if (is_begin and annotation_depth == 1)
                        else (3 + annotation_index % 2) / 4
                    )
                    bars_text.append(hovertext)
                x_ref = x

        # Specify bar parameters
        figure_height = global_max_y - global_min_y
        # Bar on full figure for memory stats figure
        if local_figure == FIG_MEMORY_STATS:
            bar_height = figure_height * (1 + bar_height_delta)
            base = global_min_y - figure_height * bar_height_delta / 2
            color = task_color
            bar_name = "Memory stats " + local_family
        # On bar per process on process figure
        else:
            bar_height = annotation_families[local_family][BAR_PERCENT]
            base = (
                -1.25 * local_process_id
                + (1 - annotation_families[local_family][BAR_PERCENT]) / 2
            )
            color = annotation_families[local_family][COLOR]
            bar_name = local_family
            # Specify bar position and label on y axis for process figure, only once
            if local_family == "Task":
                task_tickvals.append(base + 0.5)
                # add hostname to the chart if they are in the log
                if len(hostnames) > 0:
                    task_ticktext.append(
                        "Master " + hostnames[local_process_id]
                        if local_process_id == 0
                        else "Slave "
                        + str(local_process_id)
                        + " "
                        + hostnames[local_process_id]
                    )
                else:
                    task_ticktext.append(
                        "Master"
                        if local_process_id == 0
                        else "Slave " + str(local_process_id)
                    )
        # Fill missing parameters
        bars_y = []
        bars_color = []
        for i in range(len(bars_x)):
            bars_y.append(bar_height)
            bars_color.append(color)
        # Display bars
        if len(bars_x) > 0:
            fig.add_trace(
                go.Bar(
                    x=bars_x,
                    width=bars_width,
                    y=bars_y,
                    base=base,
                    marker_color=bars_color,
                    marker_opacity=bars_opacity,
                    marker_line_width=0,
                    hovertext=bars_text,
                    hoverinfo="text",
                    legendgroup=bar_name,
                    name=bar_name,
                    showlegend=bar_name not in local_process_legends,
                ),
                row=local_figure,
                col=1,
            )
            local_process_legends[bar_name] = True

    #########################################################################################
    # Main routine

    # Main parameters
    show_annotations = True  # To add annotation ticks on memory stat curves
    # To add a sub panel with the task bars per process
    show_process_figure = True
    process_figure_max_height_ratio = 0.7  # Max height ratio of task figure
    # If true, the figure height can be lower that it max height in case of few processes
    process_figure_auto_scale = False
    # IO annotations are numerous and are not shown by default, to increase readability and performance
    show_io_annotations = False
    # IO process bars could be set to False to increase performance
    show_io_process_bars = True
    # Percentage to keep in curves: all annoation are kept, by the other points can be sampled for scalability issues
    log_file_sampling_rate = 1
    # Color of task panels that are displayed on top of memory stats curves
    task_color = "lightgray"
    line_width = 1  # Width of curves

    cg = utils.ColorGenerator()

    # Huge file parameters: might be set automatically in the future
    huge_file_parameters = False
    if huge_file_parameters:
        show_annotations = False
        log_file_sampling_rate = 0.01
        show_io_process_bars = False

    # Parameters for annotations, defined by tuples for each family of annotation
    #   ANNOTATION_VISIBLE: flag to draw annotation markers on memory stats curves
    #   PROCESS_VISIBLE: flag to draw annotation sub-bars in process bars
    #   SYMBOL: marker used for annotation
    #   COLOR: color used for markers and sub-bars
    #   SIZE: size of markers
    #   BAR_PERCENT: percentage of process bar used by annotation sub-bars
    #   COMPILED_REGEX: compiled version of regex
    #   REGEX: text version of regex
    (
        ANNOTATION_VISIBLE,
        PROCESS_VISIBLE,
        SYMBOL,
        COLOR,
        SIZE,
        BAR_PERCENT,
        COMPILED_REGEX,
        REGEX,
    ) = (
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
    )  # Keys for tuple fields in annotation families
    # The annotation and subs-bars are drawn accroding to their order of declaration
    annotation_families = {}
    # Main tasks
    annotation_families["Task"] = [
        True,
        True,
        "circle",
        cg.get_color_light(),
        8,
        1.0,
        None,
        "Task.* Run [Begin|End]",
    ]
    # Master sub-routines of tasks
    annotation_families["Master"] = [
        True,
        True,
        "diamond",
        cg.get_color_light(),
        3,
        0.7,
        None,
        "Task.* "
        + "[.][ComputeResourceRequirements|MasterInitialize|MasterPrepareTaskInput|MasterAggregateResults].* [Begin|End]",
    ]
    # Slave sub-routines of tasks
    annotation_families["Slave"] = [
        True,
        True,
        "diamond",
        cg.get_color_light(),
        5,
        0.7,
        None,
        "Task.* [.][SlaveInitialize|SlaveFinalize|SlaveProcess].* [Begin|End]",
    ]
    # Anything that do not match any of the regex
    annotation_families["Other"] = [
        True,
        True,
        "triangle-up",
        cg.get_color_bold(),
        5,
        0.5,
        None,
        None,
    ]
    # Reports (all the .xls and .json files output by Khiops)
    annotation_families["Report"] = [
        True,
        True,
        "square",
        cg.get_color_bold(),
        4,
        0.5,
        None,
        ".* Write report [Begin|End]",
    ]
    # All routines related to the management of dictionaries
    annotation_families["Dictionary"] = [
        True,
        True,
        "pentagon",
        cg.get_color_bold(),
        4,
        0.5,
        None,
        "Dictionary .* [Begin|End]",
    ]
    # IO read and write basic routines
    # These annotations are numerous and are not shown by default; even their visibility in the process bars can be set
    # to False to increase the performance
    annotation_families["IO read"] = [
        show_io_annotations,
        show_io_process_bars,
        "triangle-left",
        cg.get_color_bold(),
        2,
        0.4,
        None,
        "(File .*Fill [Begin|End])|(File buffer ReadFromFile [Begin|End])",
    ]
    annotation_families["IO write"] = [
        show_io_annotations,
        show_io_process_bars,
        "triangle-right",
        cg.get_color_bold(),
        2,
        0.4,
        None,
        "(File Flush [Begin|End])|(File buffer WriteToFile [Begin|End])",
    ]

    annotation_families["Driver connect"] = [
        show_io_annotations,
        show_io_process_bars,
        "triangle-right",
        cg.get_color_bold(),
        2,
        0.1,
        None,
        "(driver.*Connect [Begin|End])",
    ]

    annotation_families["Driver ANSI"] = [
        show_io_annotations,
        show_io_process_bars,
        "triangle-right",
        cg.get_color_bold(),
        2,
        0.1,
        None,
        "(driver.*ANSI.*[Begin|End])",
    ]

    annotation_families["Driver hdfs"] = [
        show_io_annotations,
        show_io_process_bars,
        "triangle-right",
        cg.get_color_bold(),
        2,
        0.1,
        None,
        "(driver.*HDFS.*[Begin|End])",
    ]

    # Compile all regex and retrieve default family, with None regex
    none_family = None
    for family in annotation_families:
        if annotation_families[family][REGEX] is None:
            none_family = family
        else:
            annotation_families[family][COMPILED_REGEX] = re.compile(
                annotation_families[family][REGEX]
            )

    # Initialize annotation parameters, defines by tuples of three vectors for X, Y and TEXT
    (X, Y, TEXT) = (0, 1, 2)  # Keys for tuple fields in annotation parameters
    # Dictionary of annotation parameters, per (processId, family)
    annotation_parameters = {}

    # Count number of process, based on existing file with slave extensions
    process_number = utils.get_process_number(file_name)

    # Position and label of process ticks on y axis of bar figure
    task_tickvals = []
    task_ticktext = []

    # Initialize figure with two stacked subplots for memory stats curves and process bars
    FIG_MEMORY_STATS = 1  # Row index for memory stats figure
    FIG_PROCESSES = 2  # Row index for process figure
    if show_process_figure:
        # Proportion of each subplot
        if process_figure_auto_scale:
            task_proportion = min(
                max(0.05, process_figure_max_height_ratio), 0.05 * process_number
            )
        else:
            task_proportion = process_figure_max_height_ratio
        # Initialize figure with two subplots
        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.01,
            row_heights=[1 - task_proportion, task_proportion],
            row_titles=("Memory stats", "Processes"),
        )
    # Initialize one single figure for memory stats curves
    else:
        fig = make_subplots(rows=1, cols=1)

    # Host names list, at index 0 the hostname of process id O
    hostnames = []

    # Limits to countrol the max number of log lines that can be managed by curves
    max_slave_process_number = 100
    max_empty_label_line_number = 10000
    empty_label_line_sampling_frequency = None  # Will be computed once

    # Loop on files to collect curves and annotation vectors
    name_suffix = ""
    extra_width = 0
    global_min_y = 1e100
    global_max_y = -1e100

    for process_id in range(process_number):
        process_file_name = utils.build_process_filename(file_name, process_id)

        # Specific parameters per process
        if process_number > 1:
            name_suffix = (
                " Master" if process_id == 0 else " (S" + str(process_id) + ")"
            )
            extra_width = 1 if process_id == 0 else 0

        # Replace ',' with '.'
        with open(process_file_name, "r") as fProcess:
            lines = fProcess.readlines()
        with open(process_file_name, "w") as fProcess:
            for line in lines:
                fProcess.write(line.replace(",", "."))

        # Read data file
        data = pd.read_csv(process_file_name, delimiter="\t")
        assert "Time" in data, 'Variable "Time" is missing in file ' + process_file_name
        assert "Label" in data, (
            'Variable "Label" is missing in file ' + process_file_name
        )

        # Compute the ratio of lines with empty label to use, only for the main process
        if empty_label_line_sampling_frequency is None:
            line_number = len(data)
            empty_label_line_number = line_number - len(data[data.Label.notnull()])
            empty_label_line_sampling_frequency = 1
            if empty_label_line_number > max_empty_label_line_number:
                empty_label_line_sampling_frequency = (
                    empty_label_line_number // max_empty_label_line_number
                )

        # Add column to allow advance selections
        selection_column = []
        number = 0
        for ind in data.index:
            label = data["Label"][ind]
            # Case where there no label
            if not isinstance(label, str) or label == "":
                number += 1
                selection_column.append(
                    number % empty_label_line_sampling_frequency == 0
                )
            # Case of a non empty label
            else:
                # Search an index in the label (corresponds to a process index)
                # To do: case of a log related to File access
                index = -1
                label_fields = label.split(" ")
                for field in label_fields:
                    try:
                        index = int(field)
                    except ValueError:
                        index = -1
                    if index >= 0:
                        break
                # Select line if non index or index below limits
                selection = index <= max_slave_process_number
                if not selection:
                    # Sampling if not selected, with decreasing sampling rate
                    sampling_base = 1
                    while index > max_slave_process_number * sampling_base:
                        sampling_base *= 2
                        selection = index % sampling_base == 0
                selection_column.append(selection)
        data["Selection"] = selection_column

        # Filter data: max number of line with no label (memory) and max number of task processes
        filtered_data = data[data.Selection]
        if filtered_data.shape != data.shape:
            print(
                "\tFiltered data for visualization : "
                + str(filtered_data.shape)
                + " from "
                + str(data.shape)
            )
        else:
            print("\tUsed data for visualization: " + str(filtered_data.shape))
        data = filtered_data

        # Select one curve or all curves of a view
        if current_view in memory_stats_column_names:
            current_curves = [current_view]
        else:
            current_curves = view_curves[current_view]
        last_curve = current_curves[len(current_curves) - 1]

        # Draw curves of selected view
        for curve in current_curves:
            assert curve in data, (
                'Variable "' + curve + '" is missing in file ' + process_file_name
            )
            assert curve != "Time", "Variable Time is not allowed"
            assert curve != "Label", "Variable Label is not allowed"
            # Update global min and max Y
            global_min_y = min(global_min_y, min(data[curve]))
            global_max_y = max(global_max_y, max(data[curve]))
            # Extract data, with sampling if necessary
            if log_file_sampling_rate == 1:
                x_data = data["Time"]
                y_data = data[curve]
            else:
                assert 0 < log_file_sampling_rate and log_file_sampling_rate < 1, (
                    "Sampling rate "
                    + str(log_file_sampling_rate)
                    + " must be between 0 and 1"
                )
                x_data = []
                y_data = []
                for t in range(len(data["Time"])):
                    if random.random() < log_file_sampling_rate:
                        x_data.append(data["Time"][t])
                        y_data.append(data[curve][t])
            # Add curve on figure
            fig.add_trace(
                go.Scatter(
                    x=x_data,
                    y=y_data,
                    mode="lines",
                    name=curve + name_suffix,
                    legendgroup=(
                        "Labels" + str(process_id) if curve == last_curve else ""
                    ),
                    line_width=line_width + extra_width,
                )
            )

        # Collect annotations, tuples of X, Y and Text vectors
        times = data["Time"]
        # Annotations are positioned only i=n the last curve
        ylabels = data[last_curve]
        labels = data["Label"]
        # Initialize annotation parameters for the processId
        for family in annotation_families:
            annotation_parameters[process_id, family] = ([], [], [])

        # Loop on series
        for ind in data.index:
            label = labels[ind]
            # Test if the annotation is selected
            add_annotation = isinstance(label, str) and label != ""
            # Add annotation
            if add_annotation:
                # Detection of annotation family
                found_family = none_family
                for family in annotation_families:
                    compiled_regex = annotation_families[family][COMPILED_REGEX]
                    if (
                        compiled_regex is not None
                        and compiled_regex.match(label) is not None
                    ):
                        found_family = family
                        break
                # Annotation setting
                annotation_parameter = annotation_parameters[process_id, found_family]
                annotation_parameter[X].append(times[ind])
                annotation_parameter[Y].append(ylabels[ind])
                annotation_parameter[TEXT].append(label + name_suffix)

                # Hostname extraction
                if label.find("hostname") == 0:
                    start = label.index("[")
                    end = label.index("]", start)
                    hostname = label[start + 1 : end]
                    hostnames.append(hostname)

        # Add last time to show a last bar after the last task
        last_ind = data.index[-1]
        for family in annotation_families:
            annotation_parameter = annotation_parameters[process_id, family]
            # Add annotation
            annotation_parameter[X].append(times[last_ind])
            annotation_parameter[Y].append(None)
            annotation_parameter[TEXT].append(labels[last_ind])

    # Add annotation ticks on memory stat curves
    if show_annotations:
        # Loop on curves per process
        for process_id in range(process_number):
            # Add annotation per family
            for family in annotation_families:
                annotation_family = annotation_families[family]
                annotation_parameter = annotation_parameters[process_id, family]

                # Add annotation per curve
                if annotation_family[ANNOTATION_VISIBLE]:
                    fig.add_trace(
                        go.Scatter(
                            x=annotation_parameter[X],
                            y=annotation_parameter[Y],
                            mode="markers",
                            hovertext=annotation_parameter[TEXT],
                            name=family + " labels",
                            legendgroup="Labels" + str(process_id),
                            showlegend=False,
                            marker=dict(
                                symbol=annotation_family[SYMBOL],
                                color=annotation_family[COLOR],
                                size=annotation_family[SIZE] + extra_width,
                            ),
                        )
                    )

    # Global bars on memory stats figure
    process_legends = {}
    add_bars(0, "Task", FIG_MEMORY_STATS, process_legends)

    # Add bars on to show tasks
    if show_process_figure:
        # Dictionnaire des legendes deja affichees
        process_legends = {}
        # Details bars per process, and family, on the process figure
        for family in annotation_families:
            for process_id in range(process_number):
                if annotation_families[family][PROCESS_VISIBLE]:
                    add_bars(process_id, family, FIG_PROCESSES, process_legends)

    # Update axis for the memory stats figure
    yunits_label = "Allocs" if "Alloc" in current_view else "Bytes"
    fig.update_xaxes(title="Time", row=FIG_MEMORY_STATS, col=1)
    fig.update_yaxes(title=yunits_label, zeroline=True, row=FIG_MEMORY_STATS, col=1)

    # Update axis for process figure
    if show_process_figure:
        fig.update_yaxes(
            tickvals=task_tickvals,
            ticktext=task_ticktext,
            zeroline=False,
            row=FIG_PROCESSES,
            col=1,
        )
        fig.update_xaxes(title="Time", row=FIG_PROCESSES, col=1)
        fig.update_xaxes(title="", row=FIG_MEMORY_STATS, col=1)

    # Add a border for the legend
    fig.update_layout(
        legend=dict(
            bordercolor="gray",
            borderwidth=1,
            tracegroupgap=5,
            font=dict(size=12, color="black"),
        )
    )

    # Final layout
    fig.update_layout(
        title="Khiops memory log" if fig_title is None else fig_title,
        xaxis_showgrid=True,
        margin=dict(l=20, r=20, b=20, t=50),
    )

    # Draw figure
    if save_as_html:
        html_file_name = os.path.splitext(file_name)[0] + ".html"
        print("Save visualization file " + html_file_name)
        fig.write_html(html_file_name, auto_open=True)
    else:
        fig.show()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import sys
import os.path
import argparse

import _kht_constants as kht
import _kht_utils as utils
import _kht_families as test_families
import _kht_check_results as check
import kht_export

"""
Collecte de tout ou partie des resultats de test
"""

# Liste des types de collecte
COLLECT_ALL = "all"
COLLECT_WARNINGS = "warnings"
COLLECT_ERRORS = "errors"
COLLECT_TYPES = [COLLECT_ALL, COLLECT_WARNINGS, COLLECT_ERRORS]
assert len(set(COLLECT_TYPES)) == len(COLLECT_TYPES), (
    "Collect types " + str(COLLECT_TYPES) + " must not contain duplicates"
)

# Libelles des types de collecte
COLLECT_TYPE_LABELS = {
    COLLECT_ALL: "collect all test dir results",
    COLLECT_WARNINGS: "collect test dir results only in case of wanings or errors",
    COLLECT_ERRORS: "collect test dir results only in case or errors",
}
assert set(COLLECT_TYPE_LABELS) == set(
    COLLECT_TYPES
), "A label must be defined for each collect type"


def collect_learning_test_results(
    home_dir,
    input_tool_dir_name,
    input_suite_dir_name,
    input_test_dir_name,
    destination,
    family,
    collect_type,
):
    """
    Sauvegarde de l'arborescence de LearningTest
    Toute ou partie de l'arborescence source est sauvegardee selon la specification
     des operandes tool_dir_name, suite_dir_name, test_dir_name, qui peuvent etre None sinon.
    - home_dir: repertoire principal de l'aborescence source
    - tool_dir_name, suite_dir_name, test_dir_name: pour ne sauvegarder qu'une sous-partie
      de l'arborescence source si ces oprande ne sont pas None
    - destination: repertoire cible contenant l'arborescence sauvegardee
    - family: famille utilise pour choisir la sous-partie des suites a collecter
    - collect_type: pour choisir une sous-partie dees repertoires de tests a collecter
    """
    assert os.path.isdir(home_dir)
    assert family in test_families.TEST_FAMILIES
    assert collect_type in COLLECT_TYPES

    # Nom du repertoire cible
    target_home_dir = os.path.join(destination, kht.LEARNING_TEST + "_results")

    # Verification qu'il n'existe pas deja
    if os.path.isdir(target_home_dir):
        utils.fatal_error("directory " + target_home_dir + " already exists")

    # Creation du repertoire cible
    try:
        os.makedirs(target_home_dir)
    except (IOError, os.error) as why:
        utils.fatal_error(
            "unable to create target dir " + target_home_dir + " " + str(why)
        )

    # Parametrage des noms de fichiers ou repertoire specifiques a ne pas prendre en compte
    forbidden_names = ["__pycache__"]

    # Tous les outils sont a prendre en compte si on est a la racine
    if input_tool_dir_name is None:
        used_tool_names = kht.TOOL_NAMES
    # Sinon, seul l'outil correspondant au tool dir est a tester
    else:
        tool_name = kht.TOOL_NAMES_PER_DIR_NAME[input_tool_dir_name]
        used_tool_names = [tool_name]

    # Memorisation des repertoires intermediaires crees pour ne les creer que si necessaire
    created_tool_dirs = {}
    created_suite_dirs = {}

    # Parcours des repertoires des outils pour en recuperer les resultats des suites de la famille
    collected_test_dir_number = 0
    for tool_name in used_tool_names:
        tool_dir_name = kht.TOOL_DIR_NAMES[tool_name]
        # Chemin du repertoire de l'outil cible
        target_tool_dir = os.path.join(
            target_home_dir,
            tool_dir_name,
        )
        # Recherche des suites a utiliser
        if input_suite_dir_name is not None:
            assert tool_dir_name is not None
            suite_dir_names = [input_suite_dir_name]
        elif family == test_families.ALL:
            suite_dir_names = utils.sub_dirs(os.path.join(home_dir, tool_dir_name))
        else:
            suite_dir_names = test_families.FAMILY_TEST_SUITES[family, tool_name]
        # Parcours des suites de la famille
        for suite_dir_name in suite_dir_names:
            source_suite_dir = os.path.join(
                home_dir,
                tool_dir_name,
                suite_dir_name,
            )
            if os.path.isdir(source_suite_dir):
                # Chemin du repertoire de suite cible
                target_suite_dir = os.path.join(
                    target_home_dir,
                    tool_dir_name,
                    suite_dir_name,
                )
                # Repertoires de test a utiliser
                if input_test_dir_name is not None:
                    test_dir_names = [input_test_dir_name]
                else:
                    test_dir_names = os.listdir(source_suite_dir)
                # Parcours des repertoires de test de la suite
                for test_dir_name in test_dir_names:
                    source_test_dir = os.path.join(source_suite_dir, test_dir_name)
                    collect_results = False
                    if os.path.isdir(source_test_dir):
                        # On decide s'il faut collecter les results en fonction des type de collecte
                        # Cas d'une collecte systematique
                        if collect_type == COLLECT_ALL:
                            collect_results = True
                        # Cas ou cela depend des message ou erreurs du log de comparaison
                        else:
                            # Le log de comparaison doit etre disponible
                            log_file_path = os.path.join(
                                source_test_dir, kht.COMPARISON_RESULTS_LOG
                            )
                            collect_results = os.path.isfile(log_file_path)
                            # Analyse du log de comparaison s'il est disponible
                            if collect_results:
                                (
                                    error_number,
                                    warning_number,
                                    summary_infos,
                                    files_infos,
                                ) = check.analyse_comparison_log(source_test_dir)
                                # Cas de la collecte en cas de warnings
                                if collect_type == COLLECT_WARNINGS:
                                    collect_results = (
                                        error_number > 0 or warning_number > 0
                                    )
                                # Cas de la collecte d'erreurs
                                elif collect_type == COLLECT_ERRORS:
                                    collect_results = error_number > 0

                    # Copie du repertoire de test si necessaire
                    if collect_results:
                        # Creation des repertoires intermediaires uniquement si necessaire
                        if target_tool_dir not in created_tool_dirs:
                            utils.make_dir(target_tool_dir)
                            created_tool_dirs[target_tool_dir] = True
                        if target_suite_dir not in created_suite_dirs:
                            utils.make_dir(target_suite_dir)
                            created_suite_dirs[target_suite_dir] = True

                        # Copie du repertoire de test
                        target_test_dir = os.path.join(target_suite_dir, test_dir_name)
                        collected_test_dir_number += 1
                        kht_export.copy_subdirs(
                            source_test_dir,
                            target_test_dir,
                            ignore_list=forbidden_names,
                        )
            else:
                print("error : suite directory not found: " + source_suite_dir)
    print("\tCollected test dir number: " + str(collected_test_dir_number))


def main():
    def build_usage_help(
        help_command,
        help_tool_dir_name=None,
        help_suite_dir_name=None,
        help_test_dir_name=None,
        help_options=None,
    ):
        """Construction d'une ligne d'aide pour un usage de la commande test"""
        source_dir = os.path.join(".", kht.LEARNING_TEST)
        if help_test_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name, help_test_dir_name
            )
        elif help_suite_dir_name is not None:
            source_dir = os.path.join(
                source_dir, help_tool_dir_name, help_suite_dir_name
            )
        elif help_tool_dir_name is not None:
            source_dir = os.path.join(source_dir, help_tool_dir_name)
        usage_help = help_command + " " + source_dir + " (target dir)"
        if help_options is not None:
            usage_help += " " + help_options
        return usage_help

    # Nom du script
    script_file_name = os.path.basename(__file__)
    script_name = os.path.splitext(script_file_name)[0]

    # Ajout d'exemples d'utilisation
    epilog = ""
    epilog += "Usage examples"
    epilog += "\n  " + build_usage_help(script_name)
    epilog += "\n  " + build_usage_help(
        script_name, help_options="-f basic --collect-type errors"
    )
    epilog += "\n  " + build_usage_help(
        script_name,
        kht.TOOL_DIR_NAMES[kht.KHIOPS],
        help_options="--collect-type errors",
    )
    epilog += "\n  " + build_usage_help(
        script_name,
        kht.TOOL_DIR_NAMES[kht.KHIOPS],
        "Standard",
        "Iris",
    )

    # Affichage de la liste des types de collecte, en la formattant au mieux
    collect_types_help = ""
    max_collect_type_len = 0
    for collect_type in COLLECT_TYPES:
        max_collect_type_len = max(max_collect_type_len, len(collect_type))
    for collect_type in COLLECT_TYPES:
        label = COLLECT_TYPE_LABELS[collect_type]
        collect_types_help += (
            "\n  " + collect_type.ljust(max_collect_type_len + 1) + label
        )
        if collect_type == COLLECT_ALL:
            collect_types_help += " (default type)"

    # Parametrage de l'analyse de la ligne de commande
    parser = argparse.ArgumentParser(
        prog=script_name,
        description="collect a subset of a source "
        + kht.LEARNING_TEST
        + " test dirs in a target dir named "
        + kht.LEARNING_TEST
        + "_results under the destination directory",
        epilog=epilog,
        formatter_class=utils.get_formatter_class(script_name),
    )

    # Arguments positionnels
    utils.argument_parser_add_source_argument(parser)
    utils.argument_parser_add_dest_argument(parser)

    # Arguments optionnels standards
    utils.argument_parser_add_family_argument(parser)

    # Argument sur le types de collecte
    parser.add_argument(
        "--collect-type",
        help="collect type" + collect_types_help,
        choices=COLLECT_TYPES,
        default=COLLECT_ALL,
        metavar="type",
        action="store",
    )

    # Analyse de la ligne de commande
    args = parser.parse_args()

    # Verification de l'argument source
    (
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
    ) = utils.argument_parser_check_source_argument(parser, args.source)

    # Verification de l'argument destination
    utils.argument_parser_check_destination_dir(parser, home_dir, args.dest)

    # Lancement de la commande
    collect_learning_test_results(
        home_dir,
        tool_dir_name,
        suite_dir_name,
        test_dir_name,
        args.dest,
        args.family,
        args.collect_type,
    )


if __name__ == "__main__":
    utils.set_flushed_outputs()
    main()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os

import _kht_constants as kht
import _kht_utils as utils

"""
Definition des familles de test
Chaque famille de test est une liste de suite de test par outil
"""


def check_family(family):
    """Test d'existence d'une famille, avec sortie en erreur fatale si non trouvee"""
    if family not in TEST_FAMILIES:
        utils.fatal_error(
            "Family "
            + family
            + " should be in available families "
            + utils.list_to_label(TEST_FAMILIES)
        )


# Liste des familles, de la plus simple a la plus complete
BASIC = "basic"  # Famille elementaire, tres rapide a tester
FULL = "full"  # Famille correspondant a tous les tests de non regression (environ une heure a tester)
FULL_NO_KNI = "full-no-kni"  # Idem, sans KNI
COMPLETE = "complete"  # Famille complete, tres lourde a tester (environ une journee)
ALL = "all"  # Tous les repertoire de suite exhaustivement: ne pas utiliser pour les test, mais pour leur gestion
TEST_FAMILIES = [BASIC, FULL, FULL_NO_KNI, COMPLETE, ALL]
assert len(set(TEST_FAMILIES)) == len(TEST_FAMILIES), (
    "Families " + str(TEST_FAMILIES) + " must not contain duplicates"
)

# Fammile par defaut
DEFAULT_TEST_FAMILY = FULL
assert DEFAULT_TEST_FAMILY in TEST_FAMILIES

# Cas de la version V11, qui inclus des suites de test additionnelles
KHIOPS_V11 = True

# Ensuite des suites de test par famille et par outils
# Pour le cas particlier le la famille ALL, les suite ne sont pas specifiee:
# il fautparcours les sous repertoire des tool dir exhaustivement
FAMILY_TEST_SUITES = {}

# Famille basique
FAMILY_TEST_SUITES[BASIC, kht.KHIOPS] = ["Standard"]
FAMILY_TEST_SUITES[BASIC, kht.COCLUSTERING] = ["Standard"]
FAMILY_TEST_SUITES[BASIC, kht.KNI] = ["Standard"]

# Famille full
FAMILY_TEST_SUITES[FULL, kht.KHIOPS] = [
    "Standard",
    "SideEffects",
    "Rules",
    "MissingValues",
    "Advanced",
    "Bugs",
    "BugsMultiTables",
    "MultipleTargets",
    "MultiTables",
    "DeployCoclustering",
    "SparseData",
    "SparseModeling",
    "ParallelTask",
    "NewPriorV9",
    "DTClassification",
    "VariableConstruction",
    "NewV10",
    "CrashTests",
    "SmallInstability",
    "CharacterEncoding",
]
FAMILY_TEST_SUITES[FULL, kht.COCLUSTERING] = [
    "Standard",
    "Bugs",
    "NewPriorV9",
    "SmallInstability",
]
FAMILY_TEST_SUITES[FULL, kht.KNI] = ["Standard", "MultiTables", "SmallInstability"]

# Nouvelle suites specifique a la version 11
if KHIOPS_V11:
    FAMILY_TEST_SUITES[FULL, kht.KHIOPS] = FAMILY_TEST_SUITES[FULL, kht.KHIOPS] + [
        "KIInterpretation",
        "Histograms",
        "HistogramsLimits",
        "TextVariables",
    ]

# Famille full sans KNI
FAMILY_TEST_SUITES[FULL_NO_KNI, kht.KHIOPS] = FAMILY_TEST_SUITES[
    FULL, kht.KHIOPS
].copy()
FAMILY_TEST_SUITES[FULL_NO_KNI, kht.COCLUSTERING] = FAMILY_TEST_SUITES[
    FULL, kht.COCLUSTERING
].copy()
FAMILY_TEST_SUITES[FULL_NO_KNI, kht.KNI] = []

# Famille Complete
FAMILY_TEST_SUITES[COMPLETE, kht.KHIOPS] = FAMILY_TEST_SUITES[
    FULL, kht.KHIOPS
].copy() + [
    "Classification",
    "MTClassification",
    "Regression",
    "ChallengeAutoML",
]
FAMILY_TEST_SUITES[COMPLETE, kht.COCLUSTERING] = FAMILY_TEST_SUITES[
    FULL, kht.COCLUSTERING
].copy()
FAMILY_TEST_SUITES[COMPLETE, kht.KNI] = FAMILY_TEST_SUITES[FULL, kht.KNI].copy()
if KHIOPS_V11:
    FAMILY_TEST_SUITES[COMPLETE, kht.KHIOPS] = FAMILY_TEST_SUITES[
        COMPLETE, kht.KHIOPS
    ] + ["TextClassification"]



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import argparse

import _kht_constants as kht
import _kht_utils as utils


def help_commands():
    """Aide sur l'ensemble des commandes utilisateurs de LearningTest"""
    print("Available " + kht.LEARNING_TEST + " commands are")
    print(
        "- kht_test source binaries [options]"
        "\n    test a tool on a subset of test dirs"
    )
    print(
        "- kht_apply source instruction [options]"
        "\n    apply instruction (ex: errors) on a subset of test dirs"
    )
    print(
        "- kht_export source dest [options]"
        "\n    export a subset of a source "
        + kht.LEARNING_TEST
        + " tree to a target dir"
    )
    print(
        "- kht_collect_results source dest [options]"
        "\n    collect a subset of a source "
        + kht.LEARNING_TEST
        + " test dirs in a target dir"
    )
    print(
        "- kht_env"
        "\n    show the status of the main environment variables used by the tool binaries"
    )
    print("Type the name of a specific command with '-h' for detailed help")


def main():
    """Fonction principale de d'affichage de l'aide sur les commandes de LearningTest"""
    # Parametrage de l'analyse de la ligne de commande
    script_name = os.path.basename(__file__)
    base_script_name = os.path.splitext(script_name)[0]
    parser = argparse.ArgumentParser(
        prog=base_script_name,
        description="show help for all " + kht.LEARNING_TEST + " commands",
    )
    # Analyse de la ligne de commande et execution
    parser.parse_args()
    help_commands()


if __name__ == "__main__":
    utils.set_flushed_outputs()
    main()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import memory_stats_visualizer as mvs
import collect_stats_per_task as cspt
import collect_stat_io as csio

# Build a synthetic visualization of the logs and open it in a browers
# mvs.memory_stats_visualizer("./AdultSample/KhiopsMemoryStats.log", "Memory", save_as_html=True)
# mvs.memory_stats_visualizer("./AdultParallelSample/KhiopsMemoryStats.log", "Memory", save_as_html=True)

# Build a summary text file with stats per task
# cspt.collect_stats_by_task("./AdultSample/KhiopsMemoryStats.log")
# cspt.collect_stats_by_task("./AdultParallelSample/KhiopsMemoryStats.log")

# Build a summary Excel file with IO stats per task
# csio.compute_and_write_aggregates("./AdultSample/KhiopsMemoryStats.log")
# csio.compute_and_write_aggregates("./AdultParallelSample/KhiopsMemoryStats.log")



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import re
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio
import utils

pio.renderers.default = "browser"

"""
Study granted/requested and memHeap/granted ratio from memory log files
Aggregation of stats collected from memory log files
"""

# Variables of Khiops memory log file, that can be choosen for visualisation
#   Time  is used for the X axis of the curves
#   Label is used to annotated somme position on the curves and process, to eaase the interpretation
#   Each other variable can be choosen to select curves to visualize
memory_stats_column_names = [
    "Time",
    "Heap mem",
    "Max heap mem",
    "Total heap mem",
    "Alloc",
    "Max alloc",
    "Total alloc",
    "Total free",
    "Granted",
    "Max granted",
    "Total requested size",
    "Total granted size",
    "Total free size",
    "Label",
]


def memory_ratio_study(root_dir):
    """Collect memory stats from all Khiops memory log file in sub-directories of root dir
    Study granted/requested and memHeap/granted ratio from stat log files

    Unmaintained routine
    """
    # Browse sub-directories
    fig = go.Figure()
    for dir_name in os.listdir(root_dir):
        dir_path = os.path.join(root_dir, dir_name)
        if (
            "Win32" not in dir_name
            and "Adult" not in dir_name
            and "_" not in dir_name
            and os.path.isdir(dir_path)
        ):
            print(dir_path)
            all_granted = []
            all_heap_mem = []
            # Browse files per subdirectory
            for file_name in os.listdir(dir_path):
                file_path = os.path.join(dir_path, file_name)
                if file_name.find("KhiopsMemoryStats") == 0 and file_name.find(
                    ".log"
                ) == len(file_name) - len(".log"):
                    # Read data file
                    data = pd.read_csv(file_path, delimiter="\t")
                    if (
                        "Heap mem" in data
                        and "Granted" in data
                        and "Total requested size"
                        and "Total granted size" in data
                    ):
                        print("\t" + file_name)
                        data_granted = data["Granted"]
                        data_heap_mem = data["Heap mem"]
                        for i in range(len(data_granted)):
                            if i % 20 == 0:
                                all_granted.append(data_granted[i])
                                all_heap_mem.append(data_heap_mem[i])
            print("Values: " + str(len(all_granted)))
            fig.add_trace(
                go.Scatter(
                    x=all_granted,
                    y=all_heap_mem,
                    mode="markers",
                    name=dir_name,
                    marker=dict(symbol="circle", size=2),
                )
            )
    fig.show()


def collect_stats_by_task(file_name):
    """Collect memory stats per task"""

    # Defines offset collected stats per task
    (
        ID,
        SLAVES,
        START,
        TIME,
        PROCESS_NB,
        READ_TIME,
        WRITE_TIME,
        READ,
        WRITE,
        ALLOC,
        GRANTED,
        NAME,
    ) = (
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
    )  # Keys for tuple fields in annotation families
    stats_field_labels = [
        "Id",
        "Slaves",
        "Start",
        "Time",
        "Process",
        "Read T",
        "Write T",
        "Read",
        "Write",
        "GetFileSize",
        "Alloc",
        "Granted",
        "Name",
    ]
    stats_field_number = len(stats_field_labels)

    # Count number of process, based on existing file with slave extensions
    process_number = utils.get_process_number(file_name)

    # Analyse stat files for all processes to collect stats per task (and per inter-task the the master)
    all_processes_tasks = []
    for process_id in range(process_number):
        process_file_name = utils.build_process_filename(file_name, process_id)
        process_tasks = []
        all_processes_tasks.append(process_tasks)
        # Read data file
        data = pd.read_csv(process_file_name, delimiter="\t")
        data_time = data["Time"]
        data_total_alloc = data["Total alloc"]
        data_total_granted = data["Total granted size"]
        data_label = data["Label"]
        stats_size = len(data_time)
        # Collect stats per task
        inter_task_start_index = 0
        inter_task_index = 0
        total_read_time = 0
        total_write_time = 00
        total_read_number = 0
        total_write_number = 0
        total_get_file_size_number = 0
        task_name = ""
        start_index = -1
        task_process_number = 0
        start_read_index = -1
        start_write_index = -1
        for i in range(stats_size):
            label = data_label[i]
            if isinstance(label, str) and label != "":
                if "Run Begin" in label or i == stats_size - 1:
                    # Collect inter-task stats, only for process 0
                    if process_id == 0:
                        inter_task_index += 1
                        process_tasks.append(
                            [
                                "Master",
                                0,
                                data_time[inter_task_start_index],
                                data_time[i] - data_time[inter_task_start_index],
                                0,
                                total_read_time,
                                total_write_time,
                                total_read_number,
                                total_write_number,
                                total_get_file_size_number,
                                data_total_alloc[i]
                                - data_total_alloc[inter_task_start_index],
                                data_total_granted[i]
                                - data_total_granted[inter_task_start_index],
                                " Inter task " + str(inter_task_index),
                            ]
                        )
                        total_read_time = 0
                        total_write_time = 0
                        total_read_number = 0
                        total_write_number = 0
                        total_get_file_size_number = 0
                    # Initialize task stats
                    task_name = utils.extract_task_name(label)
                    start_index = i
                    task_process_number = 0
                if "Run End" in label:
                    # Collect task stats
                    assert task_name == utils.extract_task_name(label)
                    assert start_index != -1
                    process_tasks.append(
                        [
                            "Master" if process_id == 0 else "Slave" + str(process_id),
                            0 if process_id == 0 else 1,
                            data_time[start_index],
                            data_time[i] - data_time[start_index],
                            task_process_number,
                            total_read_time,
                            total_write_time,
                            total_read_number,
                            total_write_number,
                            total_get_file_size_number,
                            data_total_alloc[i]
                            - data_total_alloc[inter_task_start_index],
                            data_total_granted[i]
                            - data_total_granted[inter_task_start_index],
                            task_name,
                        ]
                    )
                    total_read_time = 0
                    total_write_time = 0
                    total_read_number = 0
                    total_write_number = 0
                    total_get_file_size_number = 0
                    # Initialize inter-task stats
                    if process_id == 0:
                        inter_task_start_index = i
                if " .SlaveProcess " in label and " Begin" in label:
                    task_process_number += 1
                if label.startswith("driver "):
                    if " fread " in label:
                        if " Begin" in label:
                            start_read_index = i
                        elif " End" in label:
                            assert start_read_index != -1
                            total_read_time += (
                                data_time[i] - data_time[start_read_index]
                            )
                            total_read_number += 1
                    elif " fwrite " in label:
                        if " Begin" in label:
                            start_write_index = i
                        elif " End" in label:
                            assert start_write_index != -1
                            total_write_time += (
                                data_time[i] - data_time[start_write_index]
                            )
                            total_write_number += 1
                    elif " GetFileSize " in label:
                        if " Begin" in label:
                            total_get_file_size_number += 1

        # Clear stats per process if no task
        if len(process_tasks) < 1:
            process_tasks.clear()
        # Collect overall stats on all the run if at least one task
        else:
            overall_stats = []
            for i in range(stats_field_number):
                overall_stats.append(0)
            overall_stats[ID] = process_tasks[0][ID]
            overall_stats[NAME] = "Overall"
            for task_stats in process_tasks:
                for i in range(stats_field_number):
                    if not isinstance(overall_stats[i], str) and not isinstance(
                        task_stats[i], str
                    ):
                        overall_stats[i] = overall_stats[i] + task_stats[i]
            overall_stats[SLAVES] = process_tasks[0][SLAVES]
            overall_stats[START] = process_tasks[0][START]
            process_tasks.append(overall_stats)

    ##########################################################
    # Collect synthetic stats over all processes

    # Initialize slave lists for each master task
    master_process_tasks = all_processes_tasks[0]
    master_process_tasks_all_slaves = []
    for _ in master_process_tasks:
        master_process_tasks_all_slaves.append([])

    # Collect slave stats for each master task
    for i, master_task in enumerate(master_process_tasks):
        # Loop on slave to collect task stats relative to master task
        for process_id in range(process_number):
            if process_id == 0:
                continue
            slave_process_tasks = all_processes_tasks[process_id]

            # Search closest master task
            best_time_distance = 1e100
            best_slave_task = None
            for slave_task in slave_process_tasks:
                if slave_task[NAME] == master_task[NAME]:
                    time_distance = abs(slave_task[START] - master_task[START])
                    if time_distance < best_time_distance:
                        best_time_distance = time_distance
                        best_slave_task = slave_task
            # Store closest slave task
            if best_slave_task is not None:
                master_process_tasks_all_slaves[i].append(best_slave_task)

    # Compute summary stats: total and mean
    if process_number > 1:
        all_processes_total_stats = []
        all_processes_mean_stats = []
        all_processes_tasks.append(all_processes_total_stats)
        all_processes_tasks.append(all_processes_mean_stats)
        for i, master_task in enumerate(master_process_tasks):
            # Initialise total with master stats
            total_stats = master_task.copy()
            all_processes_total_stats.append(total_stats)

            # Update total stats with stats collected from slaves
            all_slaves = master_process_tasks_all_slaves[i]
            for slave_task in all_slaves:
                for n in range(stats_field_number):
                    if not isinstance(total_stats[n], str):
                        total_stats[n] += slave_task[n]
            total_stats[ID] = "Total"
            total_stats[START] = master_task[START]

            # Initialise mean with master stats
            mean_stats = master_task.copy()
            all_processes_mean_stats.append(mean_stats)
            for n in range(stats_field_number):
                if not isinstance(mean_stats[n], str):
                    mean_stats[n] = 0

            # Update total stats with stats collected from slaves
            all_slaves = master_process_tasks_all_slaves[i]
            for slave_task in all_slaves:
                for n in range(stats_field_number):
                    if not isinstance(mean_stats[n], str):
                        mean_stats[n] += slave_task[n]
            for n in range(stats_field_number):
                if not isinstance(mean_stats[n], str):
                    mean_stats[n] /= max(1, len(all_slaves))
            mean_stats[ID] = "Mean(S)"
            mean_stats[START] = master_task[START]

    # Display collected stats
    stats_file_name = os.path.splitext(file_name)[0] + ".Summary.txt"
    print("Save stats summary file file " + stats_file_name)
    file_stats = open(stats_file_name, "w")
    title = ""
    for i in range(stats_field_number):
        if i > 0:
            title += "\t"
        title += stats_field_labels[i]
    # print(title)
    file_stats.write(title + "\n")
    for process_id in range(len(all_processes_tasks)):
        process_tasks = all_processes_tasks[process_id]
        for p in range(len(process_tasks)):
            task_stats = process_tasks[p]
            line = ""
            file_line = ""
            for i in range(stats_field_number):
                if i > 0:
                    line += "\t"
                    file_line += "\t"
                if isinstance(task_stats[i], str):
                    line += task_stats[i]
                elif task_stats[i] >= 100000:
                    line += "%.1E" % task_stats[i]
                elif task_stats[i] > int(task_stats[i]):
                    line += "%.1f" % task_stats[i]
                else:
                    line += str(task_stats[i])
                file_line += str(task_stats[i])
            # print(line)
            file_stats.write(file_line + "\n")
    file_stats.close()



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os.path
import sys
import stat
import warnings

import _kht_constants as kht
import _kht_utils as utils
import _kht_check_results as check
import _kht_results_management as results
import _kht_standard_instructions as standard_instructions

"""
Instruction pour des usages uniques
Peu documente et developpe rapidment sous forme de prototype
Exemples:
- manipulation a faire une fois sur l'ensemble des repertoires de test
- modification des scenario selon evoilution de l'ergonomie de Khiops core
- evaluation de l'impcat sur les performances d'une evolution des algorithme de Khiops core
- ...
"""

# Imports de pykhiops a effectuer au cas par cas dans chaque methode, car ralentissant trop les scripts
# import khiops as pk


def instruction_make_ref_time(test_dir):
    # Copie du fichier de temps vers le repertoire des resultats de reference
    results_dir = os.path.join(test_dir, kht.RESULTS)
    results_ref_dir, _ = results.get_results_ref_dir(test_dir, show=True)
    if results_ref_dir is not None:
        if not os.path.isdir(results_ref_dir):
            os.mkdir(results_ref_dir)
        if os.path.isdir(results_ref_dir):
            file_path = os.path.join(results_ref_dir, kht.TIME_LOG)
            if os.path.isfile(file_path):
                utils.remove_file(file_path)
        if os.path.isdir(results_dir) and os.path.isdir(results_ref_dir):
            utils.copy_file(
                os.path.join(results_dir, kht.TIME_LOG),
                os.path.join(results_ref_dir, kht.TIME_LOG),
            )


def instruction_make_ref_err(test_dir):
    results_dir = os.path.join(test_dir, kht.RESULTS)
    results_ref_dir, _ = results.get_results_ref_dir(test_dir, show=True)
    if results_ref_dir is not None:
        if not os.path.isdir(results_ref_dir):
            os.mkdir(results_ref_dir)
        if os.path.isdir(results_ref_dir):
            file_path = os.path.join(results_ref_dir, kht.ERR_TXT)
            utils.remove_file(file_path)
        if os.path.isdir(results_dir) and os.path.isdir(results_ref_dir):
            utils.copy_file(
                os.path.join(results_dir, kht.ERR_TXT),
                os.path.join(results_ref_dir, kht.ERR_TXT),
            )


def instruction_bench(test_dir):
    # Construction de scenario de benchmark
    def extract_info(line):
        start, end = line.split(" ", 1)
        field, comment = end.split("//", 1)
        return field

    # extraction des renseignement du fichier de parametrage
    class_file_name = ""
    class_name = ""
    database_name = ""
    target_attribute_name = ""
    prm_file_path = os.path.join(test_dir, kht.TEST_PRM)
    prm_file = open(prm_file_path, "r", errors="ignore")
    for s in prm_file:
        if s.find("class_file_name") >= 0 and class_file_name == "":
            class_file_name = extract_info(s)
        if s.find("class_name") >= 0 and class_name == "":
            class_name = extract_info(s)
        if s.find("TrainDatabase.database_name") >= 0 and database_name == "":
            database_name = extract_info(s)
        if (
            s.find("AnalysisSpec.target_attribute_name") >= 0
            and target_attribute_name == ""
        ):
            target_attribute_name = extract_info(s)
    prm_file.close()
    # affichage des lignes de fichier de bencgmark correspondant
    print("")
    print("BenchmarkSpecs.InsertItemAfter      // Insert after")
    print("// -> Benchmark")
    print("class_file_name " + class_file_name + " // Dictionary file")
    print("class_name " + class_name + " // Dictionary")
    print("target_attribute_name " + target_attribute_name + " // Target variable")
    print("database_name " + database_name + " // Database file")
    print("Exit                           // OK")
    print("// <- Benchmark")


def instruction_work(test_dir):
    results_dir = os.path.join(test_dir, kht.RESULTS)
    results_ref_dir, _ = results.get_results_ref_dir(test_dir, show=True)
    if results_ref_dir is None:
        return
    test_dir_name = utils.test_dir_name(test_dir)
    suite_dir_name = utils.suite_dir_name(test_dir)
    tool_dir_name = utils.tool_dir_name(test_dir)

    # Transformation du fichier .prm
    transform_prm = False
    if transform_prm:
        file_path = os.path.join(test_dir, kht.TEST_PRM)
        lines = utils.read_file_lines(file_path)
        try:
            with open(file_path, "w", errors="ignore") as the_file:
                for line in lines:
                    if line.find("EpsilonBinNumber") >= 0:
                        continue
                    if line.find("OutlierManagementHeuristic") >= 0:
                        continue
                    if line.find("OptimalAlgorithm") >= 0:
                        continue
                    if line.find("EpsilonBinWidth") >= 0:
                        continue
                    if line.find("MaxIntervalNumber") >= 0:
                        continue
                    if line.find("HistogramCriterion") >= 0:
                        continue
                    if line.find("MaxHierarchyLevel") >= 0:
                        continue
                    the_file.write(line)
        except Exception as e:
            print("BUG: " + file_path + " : " + str(e))

    # Parcours du repertoire de reference
    compare_histograms = True
    if compare_histograms:
        print("COMPARE " + test_dir)
        indicators = [
            "Null cost",
            "Reference null cost",
            "Cost",
            "Level",
            "Partition cost",
        ]
        if os.path.isdir(results_ref_dir):
            for file_name in os.listdir(results_ref_dir):
                ref_file_path = os.path.join(results_ref_dir, file_name)
                test_file_path = os.path.join(results_dir, file_name)
                if not os.path.isfile(test_file_path):
                    print("Missing ref file: " + file_name)
                elif "istogram.log" in file_name:
                    ref_lines = utils.read_file_lines(ref_file_path)
                    test_lines = utils.read_file_lines(test_file_path)
                    ref_indicators = {}
                    test_indicators = {}
                    ref_histogram = []
                    test_histogram = []
                    # Analyse des resultats de references
                    for line in ref_lines:
                        # Collecte des indicateurs
                        for indicator in indicators:
                            if len(line) < 70 and indicator in line:
                                fields = line[:-1].split("\t")
                                try:
                                    ref_indicators[indicator] = float(
                                        fields[len(fields) - 1]
                                    )
                                except Exception as e:
                                    print(
                                        "  "
                                        + file_name
                                        + ": Ref conversion error: "
                                        + line[:-1]
                                        + " "
                                        + str(e)
                                    )
                        # Collectes des lignes de l'histogramme
                        if (
                            len(ref_histogram) > 0
                            or "Lower bound\tUpper bound\tFrequency" in line
                        ):
                            ref_histogram.append(line)
                    # Analyse des resultats de test
                    for line in test_lines:
                        # Collecte des indicateurs
                        for indicator in indicators:
                            if len(line) < 70 and indicator in line:
                                fields = line[:-1].split("\t")
                                try:
                                    test_indicators[indicator] = float(
                                        fields[len(fields) - 1]
                                    )
                                except Exception as e:
                                    print(
                                        "  "
                                        + file_name
                                        + ": Test conversion error: "
                                        + line[:-1]
                                        + " "
                                        + str(e)
                                    )
                        # Collectes des lignes de l'histogramme
                        if (
                            len(test_histogram) > 0
                            or "Lower bound\tUpper bound\tFrequency" in line
                        ):
                            test_histogram.append(line)
                    # Comparaison des resultats
                    for indicator in indicators:
                        ref_value = ref_indicators[indicator]
                        test_value = test_indicators[indicator]
                        if (
                            abs(ref_value - test_value)
                            > abs(ref_value + test_value) / 100000
                        ):
                            print(
                                "  "
                                + file_name
                                + ": Difference in "
                                + indicator
                                + ": "
                                + str(ref_value)
                                + " vs "
                                + str(test_value)
                            )
                    if len(ref_histogram) != len(test_histogram):
                        print(
                            "  "
                            + file_name
                            + ": Difference in interval number: "
                            + str(len(ref_histogram) - 1)
                            + " vs "
                            + str(len(test_histogram) - 1)
                        )
                    else:
                        for i in range(len(ref_histogram)):
                            ref_line = ref_histogram[i]
                            test_line = test_histogram[i]
                            ref_line_fields = ref_line.split("\t")
                            test_line_fields = test_line.split("\t")
                            # Comparaison des 9 permiers champs
                            compare_ok = True
                            for f in range(8):
                                compare_ok = (
                                    compare_ok
                                    and ref_line_fields[f] == test_line_fields[f]
                                )
                                if not compare_ok:
                                    print(
                                        "  "
                                        + file_name
                                        + ": Difference in interval "
                                        + str(i)
                                        + " field "
                                        + str(f + 1)
                                        + ": \n\t"
                                        + ref_line
                                        + "\t"
                                        + test_line
                                    )
                                    break


def instruction_compare_snb_perfs(test_dir):
    """Compares the SNB performances of a test

    For each report in the results that contains a SNB predictor it compares the
    performances between the reference and the test results. The results are reported as
    a TSV table.

    The columns for each test are:
        - Family name (usually only 'TestKhiops')
        - Suite name
        - Test name
        - Subtest name: The prefix of an "*AnalysisRelsults.khj" file
        - Evaluation type: 'Test', 'Train' or 'None'
        - Metric name:
            - Classification: 'accuracy', 'auc', and 'compression'
            - Regression: 'rmse', 'mae', 'nlpd', 'rank_rmse', 'rank_mae', 'rank_nlpd'
        - Metric value
        - Reference metric value
        - Relative difference: Between 0 and 1
        - Selected variable number
        - Reference selected variable number
        - Preparation 1D flag: Empty string '' if the 1D preparations are equal,
                               'Prep1D_KO' otherwise.


    When the comparison is not feasible then the line printed will "SKIPPED" and when
    the results are not coherent with the reference these are reported as "ERROR". A
    synthetic key of the reason of failure is printed after "SKIPPED"/"ERROR".

    """
    # Import khiops-python
    try:
        from khiops import core as kh
    except ImportError:
        print("This command requires the khiops-python python package")
        exit(1)

    def _check_same_1d_preparation(prep_report_ref, prep_report):
        """Returns 'Prep1D_KO' if the 1D preparations are not equal, '' otherwise"""
        # Status values
        prep_1d_failure = "Prep1D_KO"
        prep_1d_status = ""

        # Report as Ok if both reports are not present
        if prep_report_ref is None and prep_report is None:
            pass
        # Report error if:
        # - the reference does not exist but the results do
        # - the result does not exist but the reference do
        # Similarly with the variable_statistics
        elif (
            (prep_report_ref is None and prep_report is not None)
            or (prep_report_ref is not None and prep_report is None)
            or (
                prep_report_ref.variables_statistics is None
                and prep_report.variables_statistics is not None
            )
            or (
                prep_report_ref.variables_statistics is not None
                and prep_report.variables_statistics is None
            )
        ):
            prep_1d_status = prep_1d_failure
        # Otherwise there are valid reports/var_stats in both reference and results
        else:
            # Report error if the variable statistcs differ in size
            if len(prep_report_ref.variables_statistics) != len(
                prep_report.variables_statistics
            ):
                prep_1d_status = prep_1d_failure
            # Otherwise check that each variable has the same name, level and rank
            else:
                for var_stats_ref, var_stats in zip(
                    prep_report_ref.variables_statistics,
                    prep_report.variables_statistics,
                ):
                    if (
                        var_stats_ref.name != var_stats.name
                        or var_stats_ref.rank != var_stats.rank
                        or var_stats_ref.level != var_stats.level
                    ):
                        prep_1d_status = prep_1d_failure
                        break
        return prep_1d_status

    # End of local function _check_same_1d_preparation

    # Obtain basic information of the test
    results_ref_dir, _ = results.get_results_ref_dir(test_dir, show=True)
    results_dir = os.path.join(test_dir, kht.RESULTS)
    if results_ref_dir is None:
        return
    test_dir_name = utils.test_dir_name(test_dir)
    suite_dir_name = utils.suite_dir_name(test_dir)
    tool_dir_name = utils.tool_dir_name(test_dir)

    # Obtain the sub-tests names
    khj_file_names = [
        name for name in os.listdir(results_ref_dir) if name.endswith("khj")
    ]
    sub_test_names = []
    for name in khj_file_names:
        sub_test_name = name.replace(".khj", "")
        if "AnalysisResults" in name:
            sub_test_name = sub_test_name.replace("AnalysisResults", "")
            if not sub_test_name:
                sub_test_name = "None"
        if "EvaluationReport" in name:
            sub_test_name = sub_test_name.replace("EvaluationReport", "")
            if not sub_test_name:
                sub_test_name = "None"
        sub_test_names.append(sub_test_name)

    # Ignore warnings about colliding characters in reports
    warnings.filterwarnings("ignore", message=r".*colliding")

    # Build the beginning of this test's lines
    line_prefix = f"{tool_dir_name}\t{suite_dir_name}\t{test_dir_name}"
    for file_name, sub_test_name in zip(khj_file_names, sub_test_names):
        # Read the sub-test report files
        failure_type = None
        failure_reason = None
        try:
            all_reports_ref = kh.read_analysis_results_file(
                os.path.join(results_ref_dir, file_name)
            )
        except FileNotFoundError:
            failure_type = "SKIPPED"
            failure_reason = "ReportRefNotFound"
        except kh.KhiopsJSONError as error:
            failure_type = "SKIPPED"
            failure_reason = str(error)
        if failure_reason is None:
            try:
                all_reports = kh.read_analysis_results_file(
                    os.path.join(results_dir, file_name)
                )
            except FileNotFoundError:
                failure_type = "ERROR"
                failure_reason = "ReportNotFound"
            except kh.KhiopsJSONError as error:
                failure_type = "ERROR"
                failure_reason = str(error)
        if failure_reason is None:
            if len(all_reports_ref.get_reports()) != len(all_reports.get_reports()):
                failure_type = "ERROR"
                failure_reason = "NumReportNonEq"
        # Check the type
        if failure_reason is None:
            if (
                all_reports_ref.preparation_report is not None
                and all_reports_ref.preparation_report.learning_task
                == "Unsupervised analysis"
            ):
                failure_type = "SKIPPED"
                failure_reason = "UnsupervisedTask"

        # Print the line on failure and return
        if failure_reason is not None:
            print(f"{line_prefix}\t{sub_test_name}\t{failure_type}\t{failure_reason}")
            return

        # Check that the preparation
        prep_1d_status = _check_same_1d_preparation(
            all_reports_ref.preparation_report, all_reports.preparation_report
        )

        # Get the number of selected variables
        model_report_ref = all_reports_ref.modeling_report
        model_report = all_reports.modeling_report
        if (
            model_report_ref is not None
            and "Selective Naive Bayes" in model_report_ref.get_predictor_names()
        ):
            snb_var_number_ref = (
                all_reports_ref.modeling_report.get_snb_predictor().variable_number
            )
            snb_var_number = (
                all_reports.modeling_report.get_snb_predictor().variable_number
            )
        else:
            snb_var_number_ref = -1
            snb_var_number = -1

        # Print the metric, metric_ref and the relative error of metric w/r metric_ref
        for report_ref, report in zip(
            all_reports_ref.get_reports(), all_reports.get_reports()
        ):
            if not (
                isinstance(report, kh.EvaluationReport)
                and "Selective Naive Bayes" in report.get_predictor_names()
            ):
                continue
            try:
                snb_perf_ref = report_ref.get_snb_performance()
            except ValueError as error:
                print(f"{line_prefix}\t{sub_test_name}\tERROR\tSNB: {error}")
                continue
            try:
                snb_perf = report.get_snb_performance()
            except ValueError as error:
                print(f"{line_prefix}\t{sub_test_name}\tERROR\tSNBRef: {error}")
                continue

            # Obtain the evaluation type (Train, Test or None)
            evaluation_type = report.evaluation_type
            if not evaluation_type:
                evaluation_type = "Eval"

            for metric_name in snb_perf_ref.get_metric_names():
                metric_ref = snb_perf_ref.get_metric(metric_name)
                metric = snb_perf.get_metric(metric_name)
                if metric_ref is None:
                    print(
                        f"{line_prefix}\t{sub_test_name}\t{metric_name}\t"
                        "SKIPPED\tNullRefMetric"
                    )
                elif metric is None:
                    print(
                        f"{line_prefix}\t{sub_test_name}\t{metric_name}\t"
                        "ERROR\tNullMetric"
                    )
                else:
                    if metric_ref == 0 and metric == 0:
                        rel_diff = 0
                    elif metric_ref == 0:
                        rel_diff = (metric - 1e-9) / (1e-9)
                    else:
                        rel_diff = (metric - metric_ref) / (metric_ref)
                    if 0 < abs(rel_diff) < 1e-9:
                        rel_diff = 0
                    print(
                        f"{line_prefix}\t{sub_test_name}\t{evaluation_type}\t"
                        f"{metric_name}\t{metric}\t{metric_ref}\t{rel_diff:.6g}\t"
                        f"{snb_var_number}\t{snb_var_number_ref}\t{prep_1d_status}"
                    )


def instruction_new_data_path(test_dir):
    """
    Update scenario file with new specification of data paths

    This script should work for V11 and V10 scenarios (in case of backport)
    As a "quick and dirty" script, it works well in many cases, but manual correction
    is still necessary in some cases (about ten test sets among 800 in LearningTest V11).

    The script should be applied to the whole LearningTest tree using
      "kht_apply <LarningTest dir> new-data-path -f all"
    The LearningTest directory tree should be save before using this script, in order
    to recover from some potential erroneous modifications of existing scenarios.
    """
    prm_file_path = os.path.join(test_dir, kht.TEST_PRM)
    if os.path.isfile(prm_file_path):
        # Read content of prm file
        try:
            with open(prm_file_path, "r", errors="ignore") as prm_file:
                lines = prm_file.readlines()
        except Exception as e:
            print("BUG: " + prm_file_path + " : " + str(e))
        # Update management of data path
        pattern = "DatabaseFiles.List.Key"  # Same pattern in V10 and V11
        new_lines = []
        new_scenario_necessary = False
        data_path_block_current_line = -1
        data_path_block_main_dictionary = ""
        for i, line in enumerate(lines):
            line = line.strip()
            # Case of a line containing a data path
            if pattern in line:
                new_scenario_necessary = True
                # Split line into (instruction, data path, comment)
                pos = line.find(pattern)
                #  Begin of line before data path
                instruction = line[: pos + len(pattern)]
                #  Search begin of data path
                end_line = line[pos + len(pattern) :]
                while end_line.find(" ") == 0:
                    end_line = end_line[1:]
                #  Extract datapath
                pos = end_line.find(" ")
                if pos >= 0:
                    data_path = end_line[:pos]
                    comment = end_line[pos:]
                else:
                    data_path = end_line
                    comment = ""
                if data_path.find("//") == 0:
                    comment = data_path + comment
                    data_path = ""
                debug = False
                if debug:
                    print(str(i + 1) + " :" + line)
                    print("\t" + instruction)
                    print("\t" + data_path)
                    print("\t" + comment)
                # Stop if the scenario has already been transformed
                # Allows to avoid applying the scenario several times with detrimental effects
                # But prevents the script to be applied even once in some rare cases, that
                # have to be treated manually
                if data_path == "":
                    new_scenario_necessary = False
                    break
                # Build new line
                new_line = instruction
                data_path_elements = data_path.split("`")
                assert len(data_path_elements) > 0
                # Heuristic for managing main dictionary vs external table dictionary
                if data_path_block_current_line == -1:
                    data_path_block_main_dictionary = data_path_elements[0]
                data_path_block_current_line = i
                dictionary = data_path_elements[0]
                new_line += " "
                if dictionary != data_path_block_main_dictionary:
                    new_line += "/" + dictionary
                    if len(data_path_elements) > 1:
                        new_line += "/"
                for e, element in enumerate(data_path_elements):
                    if e > 0:
                        if e > 1:
                            new_line += "/"
                        new_line += data_path_elements[e]
                new_line += comment
                new_lines.append(new_line)
                if debug:
                    print("\t" + new_line)
            # Standard case
            else:
                # Heuristic to detect a new block of data path instructions
                if i > data_path_block_current_line + 2:
                    data_path_block_current_line = -1
                    data_path_block_main_dictionary = ""
                new_lines.append(line)
        # Write content of prm file
        if new_scenario_necessary:
            try:
                with open(prm_file_path, "w", errors="ignore") as prm_file:
                    for line in new_lines:
                        prm_file.write(line + "\n")
            except Exception as e:
                print("BUG: " + prm_file_path + " : " + str(e))


def instruction_simplify_scenario_v11(test_dir):
    """
    Update scenario file (V11) to remove unnecessary default instruction for parameters APIMode and MemoryLimit
    The script should be applied to the whole LearningTest tree using
      "kht_apply <LarningTest dir> new-data-path -f all"
    The LearningTest directory tree should be save before using this script, in order
    to recover from some potential erroneous modifications of existing scenarios.
    """
    prm_file_path = os.path.join(test_dir, kht.TEST_PRM)
    if os.path.isfile(prm_file_path):
        # Read content of prm file
        try:
            with open(prm_file_path, "r", errors="ignore") as prm_file:
                lines = prm_file.readlines()
        except Exception as e:
            print("BUG: " + prm_file_path + " : " + str(e))
        # Update management of data path
        pattern1 = "AnalysisSpec.SystemParameters.APIMode true        // API mode"
        pattern2 = (
            "AnalysisSpec.SystemParameters.MemoryLimit 2000     // Memory limit in MB"
        )
        new_lines = []
        removed_lines = []
        new_scenario_necessary = False
        data_path_block_current_line = -1
        data_path_block_main_dictionary = ""
        for i, line in enumerate(lines):
            line = line.strip()
            # Case of a line containing a data path
            if pattern1 in line or pattern2 in line:
                removed_lines.append(line)
                new_scenario_necessary = True
            # Standard case
            else:
                new_lines.append(line)
                # Infos to specific instruction, to help evaluating impacts
                show_specific_instruction = False
                if show_specific_instruction:
                    instruction1 = "AnalysisSpec.SystemParameters.APIMode"
                    instruction2 = "AnalysisSpec.SystemParameters.MemoryLimit"
                    if instruction1 in line or instruction2 in line:
                        test_dir_name = utils.test_dir_name(test_dir)
                        suite_dir_name = utils.suite_dir_name(test_dir)
                        tool_dir_name = utils.tool_dir_name(test_dir)
                        # Show standard removed lines
                        if len(removed_lines) > 0:
                            for removed_line in removed_lines:
                                print(
                                    tool_dir_name
                                    + "/"
                                    + suite_dir_name
                                    + "/"
                                    + test_dir_name
                                    + ": REMOVED "
                                    + removed_line
                                )
                            removed_lines = []
                        # Show specific lines
                        print(
                            tool_dir_name
                            + "/"
                            + suite_dir_name
                            + "/"
                            + test_dir_name
                            + ": "
                            + line
                        )
                        if (
                            instruction2 in line
                            and "2000" in line
                            and not pattern2 in line
                        ):
                            print(
                                tool_dir_name
                                + "/"
                                + suite_dir_name
                                + "/"
                                + test_dir_name
                                + ": NON-STANDARD "
                                + line
                            )
        # Write content of prm file
        if new_scenario_necessary:
            try:
                with open(prm_file_path, "w", errors="ignore") as prm_file:
                    for line in new_lines:
                        prm_file.write(line + "\n")
            except Exception as e:
                print("BUG: " + prm_file_path + " : " + str(e))


def instruction_template(test_dir):
    results_dir = os.path.join(test_dir, kht.RESULTS)
    results_ref_dir, _ = results.get_results_ref_dir(test_dir, show=True)
    if results_ref_dir is None:
        return
    test_dir_name = utils.test_dir_name(test_dir)
    suite_dir_name = utils.suite_dir_name(test_dir)
    tool_dir_name = utils.tool_dir_name(test_dir)


"""
Enregistrement des instructions
"""


def register_one_shot_instructions():
    """
    Enregistrement des instructions a usage unique
    Retourne un dictionnaire d'instructions
    """

    # Gestion de l'ensemble des instructions dans un dictionnaire contenant pour chaque identifiant d'instruction
    # une paire (instruction, libelle)
    available_instructions = {}

    # Enregistrement des instructions
    standard_instructions.register_instruction(
        available_instructions,
        "makereftime",
        instruction_make_ref_time,
        "copy time file to reference results dir",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "makereferr",
        instruction_make_ref_err,
        "copy err file to reference results dir",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "bench",
        instruction_bench,
        "build bench parameter file",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "compare-snb-perfs",
        instruction_compare_snb_perfs,
        "compares the performances of all SNB predictors",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "new-data-path",
        instruction_new_data_path,
        "update scenarios with new data path spec",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "simplify-scenario-v11",
        instruction_simplify_scenario_v11,
        "update scenarios to remove default pararameters APIMode and MemoryLimit",
    )
    standard_instructions.register_instruction(
        available_instructions,
        "work",
        instruction_work,
        "last work instruction (temporary and uncommented)",
    )
    return available_instructions



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import re
import pandas as pd
import plotly.io as pio
import utils


"""
Aggregation  of the memory stats log produced by the Khiops tool

It takes log file as input and aggregate I/O time.
"""


def extract_driver_method(label):
    """Extract driver type, method Name, and Start/End"""
    driver_type = ""
    method_name = ""
    start_end = ""
    if label.find("driver") == 0:
        offset_s = label.find("[")
        offset_e = label.find("]")
        offset_be = label.find("Begin")
        if offset_be == -1:
            offset_be = label.find("End")
        assert offset_be != -1
        driver_type = label[offset_s + 1 : offset_e]
        method_name = label[offset_e + 2 : offset_be - 1]
        start_end = label[offset_be:]

        # Rename driver type in a more concise way
        schemes = ("hdfs", "ansi", "s3")
        for scheme in schemes:
            if driver_type.lower().find(scheme) != -1:
                driver_type = scheme
                break
    return driver_type, method_name, start_end


def collect_stat_io(file_name):
    """Collect memory stats per task"""
    stats = []

    # Defines offset collected stats per task
    (
        ID,
        SLAVES,
        START,
        TIME,
        PROCESS_NB,
        READ_TIME,
        WRITE_TIME,
        READ,
        WRITE,
        ALLOC,
        GRANTED,
        NAME,
    ) = (
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
    )  # Keys for tuple fields in annotation families

    # Count number of process, based on existing file with slave extensions
    process_number = utils.get_process_number(file_name)

    # Analyse stat files for all processes to collect stats per task (and per inter-task the the master)
    for process_id in range(process_number):
        process_file_name = utils.build_process_filename(file_name, process_id)

        # Read data file
        data = pd.read_csv(process_file_name, delimiter="\t")

        data_time = data["Time"]
        data_label = data["Label"]
        stats_size = len(data_time)
        start_time = 0
        for i in range(stats_size):
            label = data_label[i]

            if isinstance(label, str) and label.find("driver") == 0:
                driver_type, method_name, start_or_end = extract_driver_method(label)
                if start_or_end == "Begin":
                    assert start_time == 0
                    start_time = data_time[i]

                if start_or_end == "End":
                    duration = data_time[i] - start_time
                    start_time = 0
                    stats.append([str(process_id), driver_type, method_name, duration])
    return pd.DataFrame(stats, columns=("rank", "driver", "method", "time"))


def compute_aggregates(file_name):
    """Load data from files
    ----------

    Parameters
    ----------
    file_name: str
        name of the input memory stat log file, without suffix for the master process and
        with a suffix '_<processId>' per slave process

    Returns
    ----------
    4 DataFrames :

        - all statistics
        - statitstics grouped by method
        - statitstics grouped by method and driver
        - statitstics grouped by rank, method and driver
    """
    # Loading data
    df_stats = collect_stat_io(file_name)

    if df_stats.empty:
        raise Exception("no I/O statistics to collect")
    else:
        # Aggregation by ProcessId method and driver
        df_groupby_rank = df_stats.groupby(["rank", "method", "driver"]).agg(
            min=("time", "min"),
            max=("time", "max"),
            mean=("time", "mean"),
            median=("time", "median"),
            stddev=("time", "std"),
            sum=("time", "sum"),
            count=("time", "count"),
        )

        # Aggregation by method and driver
        df_groupby_driver = df_stats.groupby(["method", "driver"]).agg(
            min=("time", "min"),
            max=("time", "max"),
            mean=("time", "mean"),
            median=("time", "median"),
            stddev=("time", "std"),
            sum=("time", "sum"),
            count=("time", "count"),
        )

        # Aggregation by method
        df_groupby_method = df_stats.groupby(["method"]).agg(
            min=("time", "min"),
            max=("time", "max"),
            mean=("time", "mean"),
            median=("time", "median"),
            stddev=("time", "std"),
            sum=("time", "sum"),
            count=("time", "count"),
        )

        df_groupby_rank.sort_values(by=["driver", "method"], inplace=True)
        df_groupby_driver.sort_values(by=["driver", "method"], inplace=True)
        df_groupby_method.sort_values(by=["method"], inplace=True)

        return df_stats, df_groupby_method, df_groupby_driver, df_groupby_rank


def compute_and_write_aggregates(memory_log_file_name):
    """Load data from files and write results in Excel files
    ----------

    Compute and write 4 DataFrames on disk :
       - 'all_stats.xlsx' : all statistics on one file
       - 'aggregate_stats.xlsx' :
            - sheet 'group_by_method' : statitstics grouped by method
            - sheet 'group_by_driver' : statitstics grouped by method and driver
            - sheet 'group_by_rank' : rouped by rank, method and driver

    Parameters
    ----------
    memory_log_file_name: str
        The name of the input memory stat log file, without suffix for the master process and
        with a suffix '_<processId>' per slave process

    """

    # Compute all aggregates
    try:
        df_stats, df_groupby_method, df_groupby_driver, df_groupby_rank = (
            compute_aggregates(memory_log_file_name)
        )
    except Exception as error:
        print("Error in IO analysis of log file " + memory_log_file_name + ":", error)
        return

    # Writes the global statistics in a separate excel file
    dir_name = os.path.dirname(memory_log_file_name)
    file_name = os.path.basename(memory_log_file_name)
    stats_file_name = os.path.splitext(file_name)[0] + ".all_stats.xlsx"
    with pd.ExcelWriter(os.path.join(dir_name, stats_file_name)) as writer:
        df_stats.to_excel(writer)

    # Writes aggregate statistics in excel sheets
    stats_file_name = os.path.splitext(file_name)[0] + ".aggregate_stats.xlsx"
    print("Save aggregate statistics file " + stats_file_name)
    with pd.ExcelWriter(os.path.join(dir_name, stats_file_name)) as writer:
        df_groupby_method.to_excel(writer, sheet_name="group_by_method")
        df_groupby_driver.to_excel(writer, sheet_name="group_by_driver")
        df_groupby_rank.to_excel(writer, sheet_name="group_by_rank")



# Copyright (c) 2024 Orange. All rights reserved.
# This software is distributed under the BSD 3-Clause-clear License, the text of which is available
# at https://spdx.org/licenses/BSD-3-Clause-Clear.html or see the "LICENSE" file for more details.

import os
import os.path
import sys
import shutil
import stat
import argparse

import _kht_constants as kht
import _kht_families as test_families


"""
Fonction utilitaires, notamment pour la gestion des fichiers et des messages
"""

"""
Verification de la typologie des repertoires dans LearningTest
- test dir: repertoire d'un test elementaire (ex: `IrisLight`)
- suite dir: repertoire d'une famille de test, contenant un sous-repertoire par test (ex: Standard)
- tool dir: repertoire pour un outil, contenant un sous-repertoire par suite de test: (ex: TestKhiops)
- home dir: repertoire LearningTest, contenant les tool dirs, designe en externe par 'LearningTest dir'

Les methodes suivante verifie qu'un path, relatif ou absolu, se termine par un repertoire de la typologie.
En cas d'erreur, un message est affiche est on sort du programme
"""


def check_test_dir(checked_dir):
    """Test si un chemin est celui d'un repertoire de test"""
    checked_home_dir_path = parent_dir_path(checked_dir, 3)
    checked_tool_dir_name = parent_dir_name(checked_dir, 2)
    if (
        not check_home_dir(checked_home_dir_path)
        or checked_tool_dir_name not in kht.TOOL_DIR_NAMES.values()
    ):
        fatal_error(checked_dir + " should be a test directory of " + kht.LEARNING_TEST)
    return True


def check_suite_dir(checked_dir):
    """Test si un chemin est celui d'un repertoire de suite"""
    checked_home_dir_path = parent_dir_path(checked_dir, 3)
    checked_tool_dir_name = parent_dir_name(checked_dir, 1)
    if (
        not check_home_dir(checked_home_dir_path)
        or checked_tool_dir_name not in kht.TOOL_DIR_NAMES.values()
    ):
        fatal_error(
            checked_dir + " should be a suite directory of " + kht.LEARNING_TEST
        )
    return True


def check_tool_dir(checked_dir):
    """Test si un chemin est celui d'un repertoire d'outil"""
    checked_home_dir_path = parent_dir_path(checked_dir, 3)
    checked_tool_dir_name = parent_dir_name(checked_dir, 0)
    if (
        not check_home_dir(checked_home_dir_path)
        or checked_tool_dir_name not in kht.TOOL_DIR_NAMES.values()
    ):
        fatal_error(checked_dir + " should be a tool directory of " + kht.LEARNING_TEST)
    return True


def check_home_dir(checked_dir, fatal_error_if_false=True):
    """Test si un chemin est celui du repertoire LearningTest"""
    checked_home_dir_path = parent_dir_path(checked_dir, 0)
    # On n'impose pas que le repertoire racine ait le nom predefini kht.LEARNING_TEST
    # On verifie juste que le repertoire contient au moins un des repertoires d'outil
    for name in kht.TOOL_DIR_NAMES.values():
        checked_tool_dir_name = os.path.join(checked_home_dir_path, name)
        if os.path.isdir(checked_tool_dir_name):
            return True
    # Echec si aucun repertoire d'outil trouve
    if fatal_error_if_false:
        fatal_error(
            checked_dir
            + " should be a valid '"
            + kht.LEARNING_TEST
            + "' home dir, containing at least one the tools directory "
            + list_to_label(kht.TOOL_DIR_NAMES.values())
        )
    return False


def get_learning_test_sub_dir_depth(checked_dir):
    """Test si un chemin est correspond a un sous-repertoire de LearningTest
    Renvoie la profondeur a laquelle se trouver LearningTest
    - 0: home dir
    - 1: tool dir
    - 2: suite dir
    - 3: test dir
    """
    if not os.path.isdir(checked_dir):
        fatal_error(checked_dir + " should be a directory")
    checked_home_dir_path = os.path.realpath(checked_dir)
    depth = 0
    while depth < 4:
        if check_home_dir(checked_home_dir_path, fatal_error_if_false=False):
            return depth
        checked_home_dir_path = os.path.dirname(checked_home_dir_path)
        depth += 1
    fatal_error(
        checked_dir
        + " must be in a directory tree located a maximum of three levels above a valid '"
        + kht.LEARNING_TEST
        + "' home dir, containing at least one the tools directory "
        + list_to_label(kht.TOOL_DIR_NAMES.values())
    )


def get_home_dir(home_dir):
    """Retourne le repertoire de base LearningTest a partir d'un sous-repertoire de profondeur quelconque"""
    # On remonte dans le chemin (reel) jusqu'a trouver le repertoire racine
    checked_home_dir_path = os.path.realpath(home_dir)
    depth = 0
    while depth < 4:
        if check_home_dir(checked_home_dir_path, fatal_error_if_false=False):
            return checked_home_dir_path
        checked_home_dir_path = os.path.dirname(checked_home_dir_path)
        depth += 1
    assert False, (
        "No valid '" + kht.LEARNING_TEST + "' home dir found in path " + home_dir
    )


def test_dir_name(test_dir):
    """Nom du repertoire de test a partir du chemin repertoire de test"""
    return parent_dir_name(test_dir, 0)


def suite_dir_name(test_dir):
    """Nom du repertoire de suite a partir du chemin repertoire de test"""
    return parent_dir_name(test_dir, 1)


def tool_dir_name(test_dir):
    """Nom du repertoire d'outil a partir du chemin repertoire de test"""
    return parent_dir_name(test_dir, 2)


def dir_name(dir_path):
    """Renvoie le nom reel du repertoire de base d'un chemin, meme si le chemin est relatif"""
    return parent_dir_name(dir_path, 0)


def parent_dir_path(dir_path, depth):
    """Renvoie le chemin d'un repertoire parent a une profondeur donnee
    Le nom est le nom reel absolu, meme si le parametre en entree est un chemin relatif
    Ainsi, utiliser depth=0 permet d'avoir le nom reel du repertoire de base dans tous les cas

    Example: pour un test path dir_path=<root_path>/LearningTest/TestKhiops/Standard/Iris/.
    - test dir: parent_dir_name(dir_path, 0) -> Iris
    - suite dir: parent_dir_name(dir_path, 1) -> Standard
    - tool dir: parent_dir_name(dir_path, 2) -> TestKhiops
    - home dir: parent_dir_name(dir_path, 3) -> LearningTest
    """
    if not os.path.isdir(dir_path):
        fatal_error(dir_path + " should be a valid directory")
    # Recherche du parent avec un chemin relatif
    relative_parent_path = dir_path
    for i in range(depth):
        relative_parent_path += "/.."
    # Nom reel du chemin
    real_parent_path = os.path.realpath(relative_parent_path)
    return real_parent_path


def parent_dir_name(dir_path, depth):
    """Renvoie le nom d'un repertoire parent a une profondeur donnee
    Le nom est le nom reel absolu, meme si le parametre en entree est un chemin relatif
    Ainsi, utiliser depth=0 permet d'avoir le nom reel du repertoire de base dans tous les cas

    Example: pour un test path dir_path=<root_path>/LearningTest/TestKhiops/Standard/Iris/.
    - test dir: parent_dir_name(dir_path, 0) -> Iris
    - suite dir: parent_dir_name(dir_path, 1) -> Standard
    - tool dir: parent_dir_name(dir_path, 2) -> TestKhiops
    - home dir: parent_dir_name(dir_path, 3) -> LearningTest
    """
    parent_path = parent_dir_path(dir_path, depth)
    # On extrait le nom du repertoire
    result_name = os.path.basename(parent_path)
    return result_name


"""
Gestion des noms des binaires
"""


def is_valid_tool_full_exe_name(tool_exe_name):
    """Indique si le nom du binaire fait partie des noms valides, avec prise
    en compte des suffixes mpi pour les exe paralellisable"""

    return (
        tool_exe_name in kht.TOOL_EXE_NAMES.values()
        or extract_tool_exe_name(tool_exe_name) in kht.PARALLEL_TOOL_NAMES
    )


def extract_tool_exe_name(tool_full_exe_name):
    """Extrait le nom du binaire a partir d'un nom ayant potentiellement un suffixe mpi"""

    if tool_full_exe_name in kht.TOOL_EXE_NAMES.values():
        return tool_full_exe_name
    for suffix in kht.TOOL_MPI_SUFFIXES:
        if tool_full_exe_name.endswith(suffix):
            # TODO: Replace with `return tool_full_exe_name.removesuffix(suffix)`
            # as soon as Python 3.8 support is dropped
            return tool_full_exe_name[: tool_full_exe_name.index(suffix)]


"""
Gestion des messages utilisateurs
"""


def fatal_error(message):
    """Affichage du message lie a l'erreur puis quit"""
    print("error : " + message)
    exit(1)


def write_message(message, log_file=None, show=False):
    """Ecriture d'un message dans un fichier de log
    Ecriture dans un fichier de log selon le le parametre log_file
    Affichage sur la console selon le parametre show
    Si ni log_file, ni show ne sont specifier, la methode est en mode silencieux
    """
    cleaned_message = message.encode(encoding="utf-8", errors="ignore").decode(
        encoding="utf-8"
    )
    if show:
        print(cleaned_message)
    # on encode en utf-8 en ignorant les erreurs pour eviter un erreur lors de l'encodage automatique
    if log_file is not None:
        log_file.write(cleaned_message + "\n")


def append_message(initial_messages, message):
    """Ajout d'un message a un message existant, en ajoutant si necessaire ', '
     pour separer les messages si les deux sont non vides
    Retourne un message complete du nouveau message"""
    if message == "":
        return initial_messages
    elif initial_messages == "":
        return message
    else:
        return initial_messages + ", " + message


def list_to_label(value_list):
    """Creation d'un libelle a partir de l'ensemble des valeurs d'une liste
    sous la forme '(value1, value2..., valuek)'"""
    label = ""
    for value in value_list:
        if label != "":
            label += ", "
        label += str(value)
    label = "(" + label + ")"
    return label


"""
Gestion des arguments commun de la ligne de commande, pour la librairie argparse
"""


def get_formatter_class(script_name):
    """Variante de la classe de formatage a utiliser pour la creation des ArgumentParser
    - pour montrer une option par ligne, en indentant legerement l'espace des noms des options
    - pour permettre des aide multi-lignes, notamment dans l'epilogue
    (je ne sais pas pourquoi il faut pas passer par un lambda expression: pas grave)
    """
    argument_parser_formatter_class = lambda prog: argparse.RawTextHelpFormatter(
        script_name, max_help_position=26
    )
    return argument_parser_formatter_class


def argument_parser_add_source_argument(parser):
    """Ajout de l'argument source, sous repertoire dans l'arborescence LearningTest"""
    parser.add_argument(
        "source",
        help="source directory, sub-dir in a " + kht.LEARNING_TEST + " tree",
    )


def argument_parser_add_dest_argument(parser):
    """Ajout de l'argument dest, repertoire devant contenir une sous-partie d'une arborescence LearningTest"""
    parser.add_argument(
        "dest",
        help="destination directory that contains the output "
        + kht.LEARNING_TEST
        + " tree",
    )


def argument_parser_add_family_argument(parser):
    """Ajout de l'argument de famile de suites"""
    parser.add_argument(
        "-f",
        "--family",
        help="family of test suites among "
        + ", ".join(test_families.TEST_FAMILIES)
        + " (default: "
        + test_families.FULL
        + ")",
        choices=test_families.TEST_FAMILIES,
        default=test_families.FULL,
        metavar="name",
        action="store",
    )


def argument_parser_add_processes_argument(parser):
    """Ajout de l'argument du nombre de process"""
    parser.add_argument(
        "-p",
        "--processes",
        help="number of processes (default: 1)",
        dest="n",
        type=int,
        default=1,
        metavar="n",
        action="store",
    )


def argument_parser_add_forced_platform_argument(parser):
    """Ajout de l'argument de plateforme forcee en remplacement de la plateforme courante"""
    parser.add_argument(
        "--forced-platform",
        help="platform "
        + list_to_label(kht.RESULTS_REF_TYPE_VALUES[kht.PLATFORM])
        + " used to compare results (default: current platform)",
        choices=kht.RESULTS_REF_TYPE_VALUES[kht.PLATFORM],
        metavar="plt",
        action="store",
    )


def argument_parser_add_limit_test_time_arguments(parser):
    """Ajout des argument de limite des temps de test"""
    parser.add_argument(
        "--min-test-time",
        help="only for test dirs where reference test time (in file "
        + kht.TIME_LOG
        + ") is beyond a threshold",
        type=float,
        metavar="t",
        action="store",
    )
    parser.add_argument(
        "--max-test-time",
        help="only for test dirs where reference test time (in file "
        + kht.TIME_LOG
        + ") is below a threshold",
        type=float,
        metavar="t",
        action="store",
    )


def argument_parser_check_source_argument(parser, source):
    """Verification de l'argument source, a appeler apres les verification standard de parse_args()
    On renvoie la decomposition du repertoire source sous la forme des champs:
    home_dir, tool_dir_name, suite_dir_name, test_dir_name
    """
    # Verification du repertoire a tester
    learning_test_depth = get_learning_test_sub_dir_depth(source)
    source_tool_dir_name = None
    source_suite_dir_name = None
    source_test_dir_name = None
    if learning_test_depth == 3:
        source_tool_dir_name = parent_dir_name(source, 2)
        source_suite_dir_name = parent_dir_name(source, 1)
        source_test_dir_name = parent_dir_name(source, 0)
    elif learning_test_depth == 2:
        source_tool_dir_name = parent_dir_name(source, 1)
        source_suite_dir_name = parent_dir_name(source, 0)
    elif learning_test_depth == 1:
        source_tool_dir_name = parent_dir_name(source, 0)
    if (
        source_tool_dir_name is not None
        and source_tool_dir_name not in kht.TOOL_DIR_NAMES.values()
    ):
        parser.error(
            "argument source: "
            + source_tool_dir_name
            + " in "
            + os.path.realpath(source)
            + " should be a tool dir "
            + list_to_label(kht.TOOL_DIR_NAMES.values())
        )
    source_home_dir = get_home_dir(source)
    return (
        source_home_dir,
        source_tool_dir_name,
        source_suite_dir_name,
        source_test_dir_name,
    )


def argument_parser_check_destination_dir(parser, source_home_dir, destination_dir):
    """Test si un chemin peut servir a aboutir a un repertoire racine
    Il ne doit pas etre un sous-repertoire du repertoire de base"""
    check_home_dir(source_home_dir)
    home_dir = os.path.realpath(source_home_dir)
    target_dir = os.path.realpath(destination_dir)
    if (target_dir + os.path.sep).find(home_dir + os.path.sep) >= 0:
        parser.error(
            "argument dest: "
            "destination dir "
            + destination_dir
            + " must not in the directory tree "
            + home_dir
        )


def argument_parser_check_processes_argument(parser, processes):
    """Verification de l'argument processes, a appeler apres les verification standard de parse_args()"""
    max_process_number = 128
    if processes < 1:
        parser.error("argument -p/--processes: min value is 1")
    elif processes > max_process_number:
        parser.error("argument -p/--processes: max value is " + str(max_process_number))


def argument_parser_check_limit_test_time_arguments(
    parser, min_test_time, max_test_time
):
    """Verification des arguments de limites des temps de test,
    a appeler apres les verification standard de parse_args()"""
    if min_test_time is not None and min_test_time < 0:
        parser.error("argument --min-test-time must be positive")
    if max_test_time is not None and max_test_time < 0:
        parser.error("argument --max-test-time must be positive")


"""
Gestion du contenu d'un fichier
"""


def read_file_lines(file_path, log_file=None, show=False):
    """Chargement en memoire des lignes d'un fichier
    Retourne la liste des fichiers si ok, None sinon
    Ecrit un message dans le log en cas d'erreur
    """
    # lecture des lignes du fichier
    try:
        with open(file_path, "r", errors="ignore") as file:
            file_lines = file.readlines()
    except BaseException as exception:
        write_message(
            "Error : can't open file " + file_path + " (" + str(exception) + ")",
            log_file=log_file,
            show=show,
        )
        file_lines = None
    return file_lines


def write_file_lines(
    file_path, file_lines, striped_lines_suffix="\n", log_file=None, show=False
):
    """Ecriture d'une liste de ligne dans un fichier
    Ajoute un suffix aux lignes sans caractere fin de ligne
    Ecrit un message dans le log en cas d'erreur
    """
    # lecture des lignes du fichier
    try:
        with open(file_path, "w", errors="ignore") as file:
            for line in file_lines:
                file.write(line)
                if len(line) == 0 or line[-1] != "\n":
                    file.write(striped_lines_suffix)
    except BaseException as exception:
        write_message(
            "Error : can't open output file " + file_path + " (" + str(exception) + ")",
            log_file=log_file,
            show=show,
        )


"""
Gestion de patterns dans un fichier
Un patterns est une liste de sous-chaines devant se trouver en sequance dans une ligne
Exemples:
- la ligne "warning : unable to open file" contient le pattern elementaire ["warning"]
  et le pattern complexe ["warning", "file"]
"""


def find_pattern_in_line(line, pattern):
    """Renvoie la position de la premiere sous-chaine d'un pattern si une ligne contient un pattern
    Retourne -1 sinon"""
    assert isinstance(pattern, list)
    pos = 0
    first_pos = None
    for sub_pattern in pattern:
        pos = line[pos:].find(sub_pattern)
        if first_pos is None:
            first_pos = pos
        if pos == -1:
            return -1
    if first_pos is None:
        return -1
    else:
        return first_pos


def find_pattern_in_lines(lines, pattern):
    """Recherche d'un pattern un ensemble de lignes
    Renvoie l'index de la premiere ligne contenant le pattern, -1 sinon"""
    assert isinstance(lines, list)
    assert isinstance(pattern, list)
    for i, line in enumerate(lines):
        if find_pattern_in_line(line, pattern) != -1:
            return i
    return -1


def filter_lines_with_pattern(lines, pattern):
    """Retourne les lignes sans celles contenant le pattern en parametre"""
    assert isinstance(lines, list)
    assert isinstance(pattern, list)
    output_lines = []
    for line in lines:
        if find_pattern_in_line(line, pattern) == -1:
            output_lines.append(line)
    return output_lines


def filter_copyright_lines(lines):
    """Retourne les lignes sans les lignes de copyright, presentes en mode UI"""
    assert isinstance(lines, list)
    output_lines = lines
    is_copyright = False
    if len(lines) >= 2:
        is_copyright = (
            find_pattern_in_line(lines[1], ["(c)", "Orange - All rights reserved."])
            != -1
        )
    if is_copyright:
        output_lines = lines[2:]
    return output_lines


def filter_process_id_prefix_from_lines(lines):
    """Retourne les lignes sans l'eventuel prefixe de process id, du type '[0] '
    qui est emis par mpiexce dans les sorties standard"""
    output_lines = []
    for line in lines:
        # En parallelle, une ligne vide peut contenir le numero du process entre crochets
        pos_end = -1
        is_process_id = len(line) > 0 and line[0] == "["
        if is_process_id:
            pos_end = line.find("]")
            is_process_id = pos_end > 0 and line[1:pos_end].isdigit()
        if is_process_id:
            line = line[pos_end + 1 :].lstrip()
        output_lines.append(line)
    return output_lines


def filter_empty_lines(lines):
    """Retourne les lignes sans les lignes vides"""
    output_lines = []
    for line in lines:
        if line.strip() != "":
            output_lines.append(line)
    return output_lines


"""
Gestion des fichiers et repertoires
"""


def copy_file(src_file_path, dest_file_path):
    """Copie d'un fichier, avec message d'erreur"""
    try:
        shutil.copy(src_file_path, dest_file_path)
    except BaseException as message:
        print("can't copy " + src_file_path + " (" + str(message) + ")")


def remove_file(file_path):
    """Suppression d'un fichier, avec message d'erreur"""
    try:
        os.chmod(file_path, stat.S_IWRITE)
        os.remove(file_path)
    except (IOError, os.error) as why:
        print("Cannot remove file %s: %s" % (file_path, str(why)))


def make_dir(dest_dir):
    """Creation d'un repertoire, avec message d'erreur"""
    try:
        os.mkdir(dest_dir)
    except (IOError, os.error) as why:
        print("Cannot create directory %s: %s" % (dest_dir, str(why)))


def remove_dir(dir_to_remove):
    """Suppression d'un repertoire cense etre vide, avec message d'erreur"""
    try:
        os.rmdir(dir_to_remove)
    except (IOError, os.error) as why:
        print("Cannot remove directory %s: %s" % (dir_to_remove, str(why)))


def sub_dirs(source_dir):
    """Renvoie la liste des sous-repertoire d'un repertoire, sans message d'erreur"""
    result_sub_dirs = []
    if os.path.isdir(source_dir):
        try:
            list_dir = os.listdir(source_dir)
            for name in list_dir:
                if os.path.isdir(os.path.join(source_dir, name)):
                    result_sub_dirs.append(name)
        except (IOError, os.error):
            pass
    return result_sub_dirs


def set_flushed_outputs():
    """Flush systematique des sorties standard et d'erreur"""
    sys.stdout = Unbuffered(sys.stdout)
    sys.stderr = Unbuffered(sys.stderr)


class Unbuffered:
    """Pour ouvrir un fichier avec un flush systematique
    usage: par exemple, appeler sys.stdout = Unbuffered(sys.stdout) pour que toutes les sorties standard
     soit immediatement affichees dans le shell, sans bufferisation
    """

    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        # on encode en utf-8 en ignorant les erreurs pour eviter un erreur lors de l'encodage automatique
        self.stream.write(data.encode("utf-8", "ignore").decode("utf-8"))
        self.stream.flush()

    def writelines(self, datas):
        # on encode en utf-8 en ignorant les erreurs pour eviter un erreur lors de l'encodage automatique
        self.stream.writelines(
            [data.encode("utf-8", "ignore").decode("utf-8") for data in datas]
        )
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)
